"""Queue Event Config Player."""
from typing import List

from functools import partial

from mpf.core.config_player import ConfigPlayer


class QueueEventPlayer(ConfigPlayer):

    """Posts queue events based on config."""

    config_file_section = 'queue_event_player'

    __slots__ = []  # type: List[str]

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Post queue events."""
        del kwargs
        del calling_context
        if settings['events_when_finished']:
            self.machine.events.post_queue(
                settings['queue_event'],
                callback=partial(self._callback,
                                 settings['events_when_finished'],
                                 settings['args']),
                **settings['args'])
        else:
            self.machine.events.post_queue(settings['queue_event'],
                                           **settings['args'])

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        config = self._parse_config(settings, name)
        return config

    def _callback(self, event, s):
        self.machine.events.post(event, **s)

    def get_express_config(self, value):
        """No express config."""
        raise AssertionError("Not supported")



"""Contains the ModeController class for MPF."""
import importlib
import asyncio
import os
from collections import namedtuple

from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional

from mpf.core.events import QueuedEvent
from mpf.core.machine import MachineController
from mpf.core.mode import Mode
from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController

RemoteMethod = namedtuple('RemoteMethod', ['method', 'config_section', 'kwargs', 'priority'])
"""RemotedMethod is used by other modules that want to register a method to
be called on mode_start or mode_stop.

"""


class ModeController(MpfController):

    """Responsible for loading, unloading, and managing all modes in MPF."""

    config_name = "mode_controller"

    __slots__ = ["queue", "active_modes", "mode_stop_count", "_machine_mode_folders", "_mpf_mode_folders",
                 "loader_methods", "start_methods"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise mode controller.

        Args:
        ----
            machine: The main MachineController instance.

        """
        super().__init__(machine)

        # ball ending event queue
        self.queue = None                           # type: Optional[QueuedEvent]

        self.active_modes = list()                  # type: List[Mode]
        self.mode_stop_count = 0

        self._machine_mode_folders = dict()         # type: Dict[str, str]
        self._mpf_mode_folders = dict()             # type: Dict[str, str]

        # The following two lists hold namedtuples of any remote components
        # that need to be notified when a mode object is created and/or
        # started.
        self.loader_methods = list()                # type: List[RemoteMethod]
        self.start_methods = list()                 # type: List[RemoteMethod]

        if 'modes' in self.machine.config:
            # priority needs to be higher than device_manager::_load_device_modules
            self.machine.events.add_async_handler('init_phase_1', self.load_modes, priority=10)
            self.machine.events.add_handler('init_phase_2', self.initialise_modes)

        self.machine.events.add_handler('ball_ending', self._ball_ending,
                                        priority=0)

        self.machine.events.add_handler('ball_starting', self._ball_starting,
                                        priority=0)

        self.machine.events.add_handler('player_added',
                                        self._player_added, priority=0)

        self.machine.events.add_handler('player_turn_started',
                                        self._player_turn_start,
                                        priority=1000000)

        self.machine.events.add_handler('player_turn_ended',
                                        self._player_turn_ended,
                                        priority=1000000)

    def create_mode_devices(self):
        """Create mode devices."""
        for mode in self.machine.modes.values():
            mode.create_mode_devices()

    async def load_mode_devices(self):
        """Load mode devices."""
        for mode in self.machine.modes.values():
            await mode.load_mode_devices()

    def initialise_modes(self, **kwargs):
        """Initialise modes."""
        del kwargs
        # initialise modes after loading all of them to prevent races
        for item in self.loader_methods:
            for mode in self.machine.modes.values():
                if (item.config_section and
                        item.config_section in mode.config and
                        mode.config[item.config_section]):
                    item.method(config=mode.config[item.config_section],
                                mode_path=mode.path,
                                mode=mode,
                                root_config_dict=mode.config,
                                **item.kwargs)
                elif not item.config_section:
                    item.method(config=mode.config, mode_path=mode.path,
                                **item.kwargs)

        for mode in self.machine.modes.values():
            mode.initialise_mode()

    async def load_modes(self, **kwargs):
        """Load the modes from the modes: section of the machine configuration file."""
        del kwargs

        self._build_mode_folder_dicts()

        for mode in set(self.machine.config['modes']):

            if mode in self.machine.modes.values():
                raise AssertionError('Mode {} already exists. Cannot load again.'.format(mode))

            # load mode
            self.machine.modes[mode] = self._load_mode(mode)

            # add a very very short yield to prevent hangs in platforms (e.g. watchdog timeouts during IO)
            await asyncio.sleep(.0001)
            self.log.debug("Loaded mode %s", mode)

    def _find_mode_path(self, mode_string):
        asset_paths = []
        mode_path = None

        if mode_string in self._mpf_mode_folders:
            path = os.path.join(self.machine.mpf_path,
                                "modes",
                                self._mpf_mode_folders[mode_string])
            asset_paths.append(path)
            mode_path = path

        if mode_string in self._machine_mode_folders:
            path = os.path.join(self.machine.machine_path,
                                self.machine.config['mpf']['paths']['modes'],
                                self._machine_mode_folders[mode_string])

            asset_paths.append(path)
            mode_path = path

        if not mode_path:
            raise ValueError("No folder found for mode '{}'. Is your mode "
                             "folder in your machine's 'modes' folder?"
                             .format(mode_string))

        return mode_path, asset_paths

    def _get_mpf_mode_config(self, mode_string):
        try:
            mpf_mode_config = os.path.join(
                self.machine.mpf_path,
                self.machine.config['mpf']['paths']['modes'],
                self._mpf_mode_folders[mode_string],
                'config',
                self._mpf_mode_folders[mode_string] + '.yaml')
            if not os.path.isfile(mpf_mode_config):
                return False
        except KeyError:
            return False

        return mpf_mode_config

    def _get_mode_config_file(self, mode_string):
        try:
            mode_config_file = os.path.join(
                self.machine.machine_path,
                self.machine.config['mpf']['paths']['modes'],
                self._machine_mode_folders[mode_string],
                'config',
                os.path.basename(self._machine_mode_folders[mode_string]) + '.yaml')
            if not os.path.isfile(mode_config_file):
                return False
        except KeyError:
            return False
        return mode_config_file

    def _load_mode_config_spec(self, mode_string, mode_class):
        self.machine.config_validator.load_mode_config_spec(mode_string, mode_class.get_config_spec())

    def _load_mode_from_machine_folder(self, mode_string: str, code_path: str) -> Optional[Callable[..., Mode]]:
        """Load mode from machine folder and return it."""
        # this will only work for file_name.class_name
        try:
            file_name, class_name = code_path.split('.')
        except ValueError:
            return None

        # check if that mode name exist in machine folder
        if mode_string not in self._machine_mode_folders:
            return None

        # try to import
        try:
            i = importlib.import_module(
                self.machine.config['mpf']['paths']['modes'] + '.' +
                self._machine_mode_folders[mode_string] + '.code.' +
                file_name)
        except ImportError as e:
            # do not hide import error in mode
            if e.name != file_name:
                raise e
            return None

        return getattr(i, class_name, None)

    @staticmethod
    def _load_mode_from_full_path(code_path: str) -> Optional[Callable[..., Mode]]:
        """Load mode from full path.

        This is used for built-in modes like attract and game.
        """
        try:
            return Util.string_to_class(code_path)
        except ImportError as e:
            # do not hide import error in mode
            if e.name != code_path.split('.')[-1]:
                raise e
            return None

    def _load_mode_code(self, mode_string: str, code_path: str) -> Callable[..., Mode]:
        """Load code for mode."""
        # First check the machine folder
        mode_class = self._load_mode_from_machine_folder(mode_string, code_path)
        if mode_class:
            self.debug_log("Loaded code for mode %s from machine_folder", mode_string)
            return mode_class

        # load from full path
        mode_class = self._load_mode_from_full_path(code_path)
        if mode_class:
            self.debug_log("Loaded code for mode %s from full path", mode_string)
            return mode_class

        raise AssertionError("Could not load code for mode {} from {}".format(mode_string, code_path))

    def _load_mode(self, mode_string) -> Mode:
        """Load a mode, reads in its config, and creates the Mode object.

        Args:
        ----
            mode_string: String name of the mode you're loading. This is the name of
                the mode's folder in your game's machine_files/modes folder.
        """
        self.debug_log('Processing mode: %s', mode_string)

        # Find the folder for this mode. First check the machine list, and if
        # it's not there, see if there's a built-in mpf mode

        mode_path, asset_paths = self._find_mode_path(mode_string)

        config = self.machine.mpf_config.get_mode_config(mode_string)

        config['mode'] = self.machine.config_validator.validate_config("mode", config['mode'])

        # Figure out where the code is for this mode.
        if config['mode']['code']:
            # First check the machine folder
            mode_class = self._load_mode_code(mode_string, config['mode']['code'])

        else:  # no code specified, so using the default Mode class
            mode_class = Mode
            self.debug_log("Loaded default Mode() class code.")

        self._load_mode_config_spec(mode_string, mode_class)

        config['mode_settings'] = self.machine.config_validator.validate_config(
            "_mode_settings:{}".format(mode_string), config.get('mode_settings', None))

        return mode_class(self.machine, config, mode_string, mode_path, asset_paths)

    def _build_mode_folder_dicts(self):
        self._mpf_mode_folders = (
            self._get_mode_folder(self.machine.mpf_path))
        self.debug_log("Found MPF Mode folders: %s", self._mpf_mode_folders)

        self._machine_mode_folders = (
            self._get_mode_folder(self.machine.machine_path))
        self.debug_log("Found Machine-specific Mode folders: %s",
                       self._machine_mode_folders)

    def _get_mode_folder(self, base_folder):
        try:
            mode_folders = os.listdir(os.path.join(
                base_folder, self.machine.config['mpf']['paths']['modes']))
        except FileNotFoundError:
            return dict()

        final_mode_folders = dict()

        for folder in mode_folders:

            this_mode_folder = os.path.join(
                base_folder,
                self.machine.config['mpf']['paths']['modes'],
                folder)

            if os.path.isdir(this_mode_folder) and not folder.startswith('_'):
                if 'config' not in os.listdir(this_mode_folder):
                    sub_folders = self._get_sub_folders(base_folder, this_mode_folder)
                    final_mode_folders.update(sub_folders)
                else:
                    final_mode_folders[folder] = folder

        return final_mode_folders

    def _get_sub_folders(self, root_folder, base_folder):
        final_sub_folders = dict()

        for folder in os.listdir(base_folder):
            this_mode_folder = os.path.join(base_folder, folder)

            if os.path.isdir(this_mode_folder) and not folder.startswith('_'):
                if 'config' not in os.listdir(this_mode_folder):
                    final_sub_folders.update(self._get_sub_folders(root_folder, this_mode_folder))
                else:
                    final_sub_folders[folder] = os.path.relpath(this_mode_folder, os.path.join(
                        root_folder, self.machine.config['mpf']['paths']['modes']))

        return final_sub_folders

    @classmethod
    def _player_added(cls, player, num, **kwargs):
        del num
        del kwargs
        player.restart_modes_on_next_ball = list()
        '''player_var: restart_modes_on_next_ball

        desc: A list of modes that will be restarted when this player's next
        ball starts. This is more of an internal thing that MPF uses versus
        something that has a lot of value to you.
        '''

    def _player_turn_start(self, player, **kwargs):
        del kwargs
        for mode in self.machine.modes.values():
            if not mode.is_game_mode:
                continue
            mode.player = player

    def _player_turn_ended(self, player, **kwargs):
        del kwargs
        del player
        for mode in self.machine.modes.values():
            if not mode.is_game_mode:
                continue
            mode.player = None

    def _ball_starting(self, queue, **kwargs):
        del kwargs
        del queue
        for mode in self.machine.game.player.restart_modes_on_next_ball:
            self.debug_log("Restarting mode %s based on 'restart_on_next_ball"
                           "' setting", mode)

            mode.start()

        self.machine.game.player.restart_modes_on_next_ball = list()

    def _ball_ending(self, queue, **kwargs):
        """Unload all the active modes."""
        del kwargs

        if not self.active_modes:
            return

        self.queue = queue
        self.queue.wait()
        self.mode_stop_count = 0

        for mode in self.active_modes:

            if not mode.is_game_mode:
                continue

            if mode.auto_stop_on_ball_end:
                self.debug_log("Adding mode '%s' to ball ending queue", mode.name)
                self.mode_stop_count += 1
                mode.stop(callback=self._mode_stopped_callback)

            if mode.restart_on_next_ball:
                self.debug_log("Will Restart mode %s on next ball, mode")
                self.machine.game.player.restart_modes_on_next_ball.append(mode)

        if not self.mode_stop_count:
            self.queue.clear()

    def _mode_stopped_callback(self):
        self.mode_stop_count -= 1
        self.debug_log("Removing mode from ball ending queue")

        if not self.mode_stop_count:
            self.queue.clear()

    def register_load_method(self, load_method, config_section_name=None,
                             priority=0, **kwargs):
        """Register a method which is called when the mode is loaded.

        Used by core components, plugins, etc. to register themselves with
        the Mode Controller for anything they need a mode to do when it's
        registered.

        Args:
        ----
            load_method: The method that will be called when this mode code
                loads.
            config_section_name: An optional string for the section of the
                configuration file that will be passed to the load_method when
                it's called.
            priority: Int of the relative priority which allows remote methods
                to be called in a specific order. Default is 0. Higher values
                will be called first.
            **kwargs: Any additional keyword arguments specified will be passed
                to the load_method.

        Note that these methods will be called once, when the mode code is first
        initialized during the MPF boot process.
        """
        if not callable(load_method):
            raise ValueError("Cannot add load method '{}' as it is not"
                             "callable".format(load_method))

        self.loader_methods.append(RemoteMethod(method=load_method,
                                                config_section=config_section_name, kwargs=kwargs,
                                                priority=priority))
        self.loader_methods.sort(key=lambda x: x.priority, reverse=True)

    def register_start_method(self, start_method, config_section_name=None,
                              priority=0, **kwargs):
        """Register a method which is called anytime a mode is started.

        Used by core components, plugins, etc. to register themselves with
        the Mode Controller for anything that they a mode to do when it starts.

        Args:
        ----
            start_method: The method that will be called when this mode code
                loads.
            config_section_name: An optional string for the section of the
                configuration file that will be passed to the start_method when
                it's called.
            priority: Int of the relative priority which allows remote methods
                to be called in a specific order. Default is 0. Higher values
                will be called first.
            **kwargs: Any additional keyword arguments specified will be passed
                to the start_method.

        """
        if not callable(start_method):
            raise ValueError("Cannot add start method '{}' as it is not"
                             "callable".format(start_method))

        self.debug_log('Registering %s as a mode start method. Config '
                       'section: %s, priority: %s, kwargs: %s',
                       start_method, config_section_name, priority, kwargs)

        self.start_methods.append(RemoteMethod(method=start_method,
                                               config_section=config_section_name, priority=priority,
                                               kwargs=kwargs))

        self.start_methods.sort(key=lambda x: x.priority, reverse=True)

    def remove_start_method(self, start_method, config_section_name=None, priority=0, **kwargs):
        """Remove an existing start method."""
        method = RemoteMethod(method=start_method, config_section=config_section_name,
                              priority=priority, kwargs=kwargs)

        if method in self.start_methods:
            self.start_methods.remove(method)

    def set_mode_state(self, mode: Mode, active: bool):
        """Remember mode state."""
        if active:
            self.active_modes.append(mode)
        else:
            self.active_modes.remove(mode)

        # sort the active mode list by priority
        self.active_modes.sort(key=lambda x: (x.priority, x.name), reverse=True)

        # notify about changed active mode list
        self.machine.events.post("modes_active_modes_changed")
        '''event: modes_active_modes_changed

        The list of active mode changed. A mode has been started or stopped.
        '''

        if self._debug:
            self.dump()

    def dump(self):
        """Dump the current status of the running modes to the log file."""
        self.debug_log('+=========== ACTIVE MODES ============+')

        for mode in self.active_modes:
            if mode.active:
                self.debug_log('| {} : {}'.format(
                    mode.name, mode.priority).ljust(38) + '|')

        self.debug_log('+-------------------------------------+')

    def is_active(self, mode_name) -> bool:
        """Return true if the mode is active, False if it is not.

        Args:
        ----
            mode_name: String name of the mode to check.
        """
        return mode_name in [x.name for x in self.active_modes
                             if x.active is True]



"""Contains the BallLock device class."""
from typing import List, Optional

from mpf.core.enable_disable_mixin import EnableDisableMixin

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.mode_device import ModeDevice

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.playfield import Playfield     # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("locked_balls")
class MultiballLock(EnableDisableMixin, ModeDevice):

    """Ball lock device which locks balls for a multiball."""

    config_section = 'multiball_locks'
    collection = 'multiball_locks'
    class_label = 'multiball_lock'

    __slots__ = ["lock_devices", "source_playfield", "_events", "_locked_balls", "_source_devices"]

    def __init__(self, machine, name):
        """Initialise ball lock."""
        self.lock_devices = []
        self.source_playfield = None    # type: Optional[Playfield]
        self._source_devices = None     # type: Optional[List[BallDevice]]

        # initialise variables
        self._events = {}

        self._locked_balls = 0
        # Locked balls in case we are keep_virtual_ball_count_per_player is false

        super().__init__(machine, name)

    async def _initialize(self):
        # load lock_devices
        await super()._initialize()

        self.lock_devices = []
        for device in self.config['lock_devices']:
            self.lock_devices.append(device)
            self._events[device] = []

        self.source_playfield = self.config['source_playfield']
        self._source_devices = self.config['source_devices']

        self.machine.events.add_handler("player_turn_starting", self._player_turn_starting)

    def _enable(self):
        """Enable the lock.

        If the lock is not enabled, no balls will be locked.
        """
        self.debug_log("Enabling...")
        self._register_handlers()

    def _player_turn_starting(self, queue, **kwargs):
        del kwargs
        # reset locked balls
        self._locked_balls = 0

        # check if the lock is physically full and not virtually full and release balls in that case
        if self._physically_remaining_space <= 0 and not self.is_virtually_full:
            self.log.info("Will release a ball because the lock is physically full but not virtually for the player.")
            # TODO: eject to next playfield
            self.lock_devices[0].eject()
            queue.wait()
            self.machine.events.add_handler("ball_drain", self._wait_for_drain, queue=queue)

    def _wait_for_drain(self, queue, balls, **kwargs):
        del kwargs
        if balls <= 0:
            return {'balls': balls}

        self.debug_log("Ball of lock drained.")

        queue.clear()

        self.machine.events.remove_handler_by_event('ball_drain', self._wait_for_drain)

        return {'balls': balls - 1}

    def _disable(self):
        """Disable the lock.

        If the lock is not enabled, no balls will be locked.
        """
        self.debug_log("Disabling...")
        self._unregister_handlers()

    @event_handler(1)
    def event_reset_all_counts(self, **kwargs):
        """Event handler for reset_all_counts event."""
        del kwargs
        self.reset_all_counts()

    def reset_all_counts(self):
        """Reset the locked balls for all players."""
        if self.config['locked_ball_counting_strategy'] not in ("virtual_only", "min_virtual_physical"):
            raise AssertionError("Count is only tracked per player")
        for player in self.machine.game.player_list:
            player['{}_locked_balls'.format(self.name)] = 0

    @event_handler(2)
    def event_reset_count_for_current_player(self, **kwargs):
        """Event handler for reset_count_for_current_player event."""
        del kwargs
        self.reset_count_for_current_player()

    def reset_count_for_current_player(self):
        """Reset the locked balls for the current player."""
        if self.config['locked_ball_counting_strategy'] in ("virtual_only", "min_virtual_physical", "no_virtual"):
            self.locked_balls = 0
        else:
            raise AssertionError("Cannot reset physical balls")

    @property
    def locked_balls(self):
        """Return the number of locked balls for the current player."""
        if not self.machine.game:
            # this is required for the monitor because it will query this variable outside of a game
            # remove when #893 is fixed
            return None

        if self.config['locked_ball_counting_strategy'] == "virtual_only":
            return self.machine.game.player['{}_locked_balls'.format(self.name)]
        if self.config['locked_ball_counting_strategy'] == "min_virtual_physical":
            return min(self.machine.game.player['{}_locked_balls'.format(self.name)], self._physically_locked_balls)
        if self.config['locked_ball_counting_strategy'] == "physical_only":
            return self._physically_locked_balls

        return self._locked_balls

    @locked_balls.setter
    def locked_balls(self, value):
        """Set the number of locked balls for the current player."""
        if self.config['locked_ball_counting_strategy'] in ("virtual_only", "min_virtual_physical"):
            self.machine.game.player['{}_locked_balls'.format(self.name)] = value
        elif self.config['locked_ball_counting_strategy'] in "no_virtual":
            self._locked_balls = value
        else:
            raise AssertionError("Cannot write locked_balls for strategy {}".format(
                self.config['locked_ball_counting_strategy']))

    def _register_handlers(self):
        priority = (self.mode.priority if self.mode else 0) + \
            self.config['priority']
        # register on ball_enter of lock_devices
        for device in self.lock_devices:
            self.machine.events.add_handler(
                'balldevice_' + device.name + '_ball_enter',
                self._lock_ball, device=device, priority=priority)
            self.machine.events.add_handler(
                'balldevice_' + device.name + '_ball_entered',
                self._post_events, device=device, priority=priority)

    def _unregister_handlers(self):
        # unregister ball_enter handlers
        self.machine.events.remove_handler(self._lock_ball)
        self.machine.events.remove_handler(self._post_events)

    @property
    def is_virtually_full(self):
        """Return true if lock is full."""
        return self.remaining_virtual_space_in_lock <= 0

    @property
    def remaining_virtual_space_in_lock(self):
        """Return the remaining capacity of the lock."""
        balls = self.config['balls_to_lock'] - self.locked_balls
        if balls < 0:
            balls = 0
        return balls

    @property
    def _max_balls_locked_by_any_player(self):
        """Return the highest number of balls locked for all players."""
        max_balls = 0
        for player in self.machine.game.player_list:
            if max_balls < player['{}_locked_balls'.format(self.name)]:
                max_balls = player['{}_locked_balls'.format(self.name)]

        return max_balls

    @property
    def _physically_locked_balls(self):
        """Return the number of physically locked balls."""
        balls = 0
        for device in self.lock_devices:
            balls += device.available_balls

        return balls

    @property
    def _physically_remaining_space(self):
        """Return the space in the physically locks."""
        balls = 0
        for device in self.lock_devices:
            balls += device.capacity - device.available_balls

        return balls

    def _lock_ball(self, unclaimed_balls: int, new_available_balls: int, device: "BallDevice", **kwargs):
        """Handle result of the _ball_enter event of lock_devices."""
        del kwargs
        # if there are no balls do not claim anything
        if unclaimed_balls <= 0:
            return {'unclaimed_balls': unclaimed_balls}

        # MPF will make sure that devices get one event per ball
        assert unclaimed_balls == 1

        if not self.machine.game or not self.machine.game.player:
            # bail out if we are outside of a game
            return {'unclaimed_balls': unclaimed_balls}

        # if already full do not take any balls
        if self.is_virtually_full:
            self.debug_log("Cannot lock balls. Lock is full.")
            return {'unclaimed_balls': unclaimed_balls}

        # first take care of virtual ball count in lock
        capacity = self.remaining_virtual_space_in_lock
        # take ball up to capacity limit
        if unclaimed_balls > capacity:
            balls_to_lock = capacity
        else:
            balls_to_lock = unclaimed_balls

        new_locked_balls = self.locked_balls + 1
        # post event for ball capture
        self._events[device].append({"event": 'multiball_lock_' + self.name + '_locked_ball',
                                     "total_balls_locked": new_locked_balls})
        '''event: multiball_lock_(name)_locked_ball
        desc: The multiball lock device (name) has just locked one additional ball.

        args:
            total_balls_locked: The current total number of balls this device
                has locked.
        '''
        if self.config['locked_ball_counting_strategy'] != "physical_only":
            self.locked_balls = new_locked_balls

        # now check how many balls we want physically in the lock
        balls_to_lock_physically = balls_to_lock

        if self._physically_remaining_space < new_available_balls:
            # we cannot lock if there isn't any space left
            balls_to_lock_physically = 0
            self.debug_log("Will not keep the ball. Device is full. Remaining space: %s. Balls to lock: %s",
                           self._physically_remaining_space, balls_to_lock)

        if self.config['locked_ball_counting_strategy'] in ("virtual_only", "min_virtual_physical"):
            # only keep ball if any player could use it
            if self._max_balls_locked_by_any_player < self._physically_locked_balls + new_available_balls:
                self.debug_log("Will not keep ball because no player could use it. Max locked balls by any player "
                               "is %s and we physically got %s", self._max_balls_locked_by_any_player,
                               self._physically_locked_balls)
                balls_to_lock_physically = 0

        if self.config['locked_ball_counting_strategy'] == "min_virtual_physical":
            # do not lock if the lock would be physically full but not virtually
            if (self._physically_remaining_space <= new_available_balls and
                    self.config['balls_to_lock'] - self.machine.game.player['{}_locked_balls'.format(self.name)] > 0):
                self.debug_log("Will not keep ball because the lock would be physically full but virtually still "
                               "has space for this player.")
                balls_to_lock_physically = 0
        elif self.config['locked_ball_counting_strategy'] != "physical_only":
            # do not lock if the lock would be physically full but not virtually
            if not self.is_virtually_full and self._physically_remaining_space <= new_available_balls:
                balls_to_lock_physically = 0
                self.debug_log("Will not keep ball because the lock would be physically full but virtually still "
                               "has space for this player.")

        # check if we are full now and post event if yes
        if (self.config['locked_ball_counting_strategy'] == "physical_only" and
            new_locked_balls >= self.config['balls_to_lock']) or \
                self.remaining_virtual_space_in_lock == 0:
            self._events[device].append({'event': 'multiball_lock_' + self.name + '_full',
                                         'balls': new_locked_balls})
        '''event: multiball_lock_(name)_full
        desc: The multiball lock device (name) is now full.
        args:
            balls: The number of balls currently locked in this device.
        '''

        # schedule eject of new balls for all physically locked balls
        if self.config['balls_to_replace'] == -1 or self.locked_balls <= self.config['balls_to_replace']:
            self.debug_log("{} locked balls and {} to replace, requesting {} new balls"
                           .format(self.locked_balls, self.config['balls_to_replace'], balls_to_lock_physically))
            self._request_new_balls(balls_to_lock_physically)
        else:
            self.debug_log("{} locked balls exceeds {} to replace, not requesting any balls"
                           .format(self.locked_balls, self.config['balls_to_replace']))

        self.debug_log("Locked %s balls virtually and %s balls physically", balls_to_lock, balls_to_lock_physically)

        return {'unclaimed_balls': unclaimed_balls - balls_to_lock_physically}

    def _post_events(self, device, **kwargs):
        """Post events on callback from _ball_entered handler.

        Events are delayed to this handler because we want the ball device to have accounted for the balls.
        """
        del kwargs
        for event in self._events[device]:
            self.machine.events.post(**event)
        self._events[device] = []

    def _request_new_balls(self, balls):
        """Request new ball to playfield."""
        balls_added = 0
        for device in self._source_devices:
            balls_to_add = max(min(device.available_balls, balls - balls_added), 0)
            device.eject(balls=balls_to_add, target=self.source_playfield)
            balls_added += balls_to_add

        self.source_playfield.add_ball(balls=max(balls - balls_added, 0))



"""A show queue which can will be played sequentially."""
from collections import deque

from mpf.assets.show import RunningShow, ShowConfig
from mpf.core.system_wide_device import SystemWideDevice

MYPY = False
if MYPY:   # pragma: no cover
    from typing import Tuple, Deque, Optional


class ShowQueue(SystemWideDevice):

    """Represents a show queue."""

    config_section = 'show_queues'
    collection = 'show_queues'
    class_label = 'show_queue'

    __slots__ = ["shows_queue", "_current_show"]

    def __init__(self, machine, name):
        """Initialise show queue."""
        super().__init__(machine, name)

        self.shows_queue = deque()  # type: Deque[Tuple[ShowConfig, int]]
        self._current_show = None   # type: Optional[RunningShow]

    def enqueue_show(self, show_config: ShowConfig, start_step: int):
        """Add a show to the end of the queue."""
        self.shows_queue.append((show_config, start_step))
        if not self._current_show:
            self._play_next_show()

    def _play_next_show(self):
        """Play the next show."""
        if not self.shows_queue:
            # no show queued
            self._current_show = None
            return

        show_config, start_step = self.shows_queue.popleft()
        self._current_show = self.machine.show_controller.replace_or_advance_show(self._current_show, show_config,
                                                                                  start_step=start_step,
                                                                                  stop_callback=self._play_next_show)



"""Boards and wiring strategy for FAST Pinball."""
from typing import List, Dict, Optional, TypeVar, Any, Callable, Tuple
from mpf.core.machine import MachineController
from mpf.wire.base import Board, Pin, System, PinSpecs, SerialLED, Switch, Coil


class FastPFB(Board):

    """The FAST Power Filter Board."""

    def __init__(self):
        """Set up a power filter board."""
        super().__init__("PFB")
        self.add_connector("J1", [    # Coin door switch
            ("SW", 5),
            ("KEY", -1),
            ("GND", 0)
        ])
        self.add_connector("J2", [    # Coil power enable
            ("+", 5),
            ("-", 0),
            ("KEY", -1)
        ])
        self.add_connector("J3", [    # PSU In
            ("HV+", 48),
            ("HV+", 48),
            ("GND", 0),
            ("GND", 0),
            ("5v", 5),
            ("5v", 5),
            ("12v", 12),
            ("KEY", -1),
            ("GND", 0),
            ("GND", 0),
            ("V1", -2),
            ("V2", -2)
        ])
        self.add_connector("J4", [    # Power Out
            ("HV+", 48),
            ("HV+", 48),
            ("GND", 0),
            ("GND", 0),
            ("5v", 5),
            ("5v", 5),
            ("KEY", -1),
            ("12v", 12),
            ("GND", 0),
            ("GND", 0),
            ("V1", -2),
            ("V2", -2)
        ])


class FastNetBoard(Board):

    """A generic class for boards with FAST networking connectors."""

    @staticmethod
    def fast_net_connector() -> PinSpecs:
        """Return the pin spec to use for a FAST network connector."""
        return [("RJ45", 1)]

    def get_net_in(self) -> Pin:
        """Return this board's network in 'pin' (actually a whole connector with irrelevant pinout)."""

    def get_net_out(self) -> Pin:
        """Return this board's network out 'pin''."""


class FastIOBoard(FastNetBoard):

    """Generic class for FAST IO boards."""

    @staticmethod
    # pylint: disable=too-many-arguments
    def fast_connector_block(prefix: str, size: int, vclass: int, offset: int, key_loc: int, first_ground: int) \
            -> PinSpecs:
        """Return the pinspec for a FAST IO connector.

        :param prefix The name to place before the name of each pin.
        :param size The number of pins on the connector.
        :param vclass The voltage class of pins on the connector.
        :param offset The number of the first switch/driver on the connector.
        :param key_loc The index of the key pin.
        :param first_ground The index of the first ground pin.
        """
        pins = []
        non_key_pins = 0
        for pin in range(size):
            if pin == key_loc:
                pins.append(("KEY", -1))
            elif pin >= first_ground:
                pins.append(("GND", 0))
            else:
                pins.append((prefix + (str(non_key_pins + offset)), vclass))
                non_key_pins += 1
        return pins

    def get_switch_pin(self, switch_id) -> Pin:
        """Return the pin for the numbered switch."""

    def get_driver_pin(self, driver_id) -> Pin:
        """Return the pin for the numbered driver."""

    def get_switch_grounds(self, index) -> List[Pin]:
        """Return a list of all index'th switch ground pins."""

    def get_driver_grounds(self, index) -> List[Pin]:
        """Return a list of all index'th driver ground pins."""


class FastIO3208(FastIOBoard):

    """A Fast IO 3028 board with 32 switches and 8 drivers."""

    def __init__(self):
        """Initialize the board."""
        super().__init__("IO 3208")
        self.add_connector("J1", self.fast_net_connector())
        self.add_connector("J2", self.fast_net_connector())
        self.add_connector("J3", self.fast_connector_block("Sw", 11, 5, 8, 3, 9))
        self.add_connector("J4", self.fast_connector_block("Dr", 12, 48, 0, 5, 9))
        self.add_connector("J6", self.fast_connector_block("Sw", 11, 5, 16, 2, 9))
        self.add_connector("J8", self.fast_connector_block("Sw", 11, 5, 0, 4, 9))
        self.add_connector("J9", self.fast_connector_block("Sw", 11, 5, 24, 1, 9))

    def get_net_out(self) -> Pin:
        """Return the network out port, J1."""
        return self["J1"][0]

    def get_net_in(self) -> Pin:
        """Return the network out port, J0."""
        return self["J2"][0]

    # pylint: disable=too-many-return-statements
    def get_switch_pin(self, switch_id) -> Pin:
        """Return the pin for the numbered switch."""
        if switch_id < 4:
            return self["J8"][switch_id]
        if switch_id < 8:
            return self["J8"][switch_id + 1]
        if switch_id < 11:
            return self["J3"][switch_id - 8]
        if switch_id < 16:
            return self["J3"][switch_id - 7]
        if switch_id < 18:
            return self["J6"][switch_id - 16]
        if switch_id < 24:
            return self["J6"][switch_id - 15]
        if switch_id == 24:
            return self["J9"][0]
        if switch_id < 32:
            return self["J9"][switch_id - 23]
        return None

    def get_driver_pin(self, driver_id) -> Pin:
        """Return the pin for the numbered driver."""
        if driver_id < 5:
            return self["J4"][driver_id]
        if driver_id < 8:
            return self["J4"][driver_id + 1]
        return None

    def get_switch_grounds(self, index: int) -> List[Pin]:
        """Get the index'th switch ground pin from each of 4 connectors."""
        return [self["J3"][9 + index],
                self["J6"][9 + index],
                self["J8"][9 + index],
                self["J9"][9 + index]]

    def get_driver_grounds(self, index: int) -> List[Pin]:
        """Get the index'th driver ground pin for the connector."""
        return [self["J4"][9 + index]]


class FastIO1616(FastIOBoard):

    """A FAST IO 1616 board with 16 switches and 16 drivers."""

    def __init__(self):
        """Initialize the board."""
        super().__init__("IO 1616")
        self.add_connector("J1", self.fast_net_connector())
        self.add_connector("J2", self.fast_net_connector())
        self.add_connector("J3", self.fast_connector_block("Dr", 12, 48, 0, 4, 9))
        self.add_connector("J4", self.fast_connector_block("Dr", 12, 48, 8, 5, 9))
        self.add_connector("J7", self.fast_connector_block("Sw", 11, 5, 0, 4, 9))
        self.add_connector("J8", self.fast_connector_block("Sw", 11, 5, 8, 3, 9))

    def get_net_out(self):
        """Get the network output port, J1."""
        return self["J1"][0]

    def get_net_in(self):
        """Get the network input port, J2."""
        return self["J2"][0]

    def get_switch_grounds(self, index):
        """Get the index'th switch grounds for the two switch connectors."""
        return [self["J7"][9 + index],
                self["J8"][9 + index]]

    def get_driver_grounds(self, index):
        """Get the index'th driver grounds for the two driver connectors."""
        return [self["J3"][9 + index],
                self["J4"][9 + index]]

    def get_switch_pin(self, switch_id):
        """Return the pin for the numbered switch."""
        if switch_id < 4:
            return self["J7"][switch_id]
        if switch_id < 8:
            return self["J7"][switch_id + 1]
        if switch_id < 11:
            return self["J8"][switch_id - 8]
        if switch_id < 16:
            return self["J8"][switch_id - 7]
        return None

    def get_driver_pin(self, driver_id):
        """Return the pin for the numbered driver."""
        if driver_id < 4:
            return self["J3"][driver_id]
        if driver_id < 8:
            return self["J3"][driver_id + 1]
        if driver_id < 13:
            return self["J4"][driver_id - 8]
        if driver_id < 16:
            return self["J4"][driver_id - 7]
        return None


class FastIO0804(FastIOBoard):

    """The FAST IO 0804 board with 8 switches and 4 drivers."""

    def __init__(self):
        """Initialize the board."""
        super().__init__("IO 0807")
        self.add_connector("J1", self.fast_net_connector())
        self.add_connector("J2", self.fast_net_connector())
        self.add_connector("J3", self.fast_connector_block("Dr", 7, 48, 0, 4, 5))
        self.add_connector("J4", self.fast_connector_block("SW", 11, 5, 0, 4, 9))

    def get_net_out(self):
        """Get the network output port, J1."""
        return self["J1"][0]

    def get_net_in(self):
        """Get the network input port, J2."""
        return self["J2"][0]

    def get_switch_pin(self, switch_id):
        """Return the pin for the numbered switch."""
        if switch_id < 4:
            return self["J4"][switch_id]
        if switch_id < 8:
            return self["J4"][switch_id + 1]
        return None

    def get_driver_pin(self, driver_id):
        """Return the pin for the numbered driver."""
        assert driver_id <= 4
        return self["J3"][driver_id]

    def get_switch_grounds(self, index):
        """Return the index'th switch ground pin for the single connector."""
        return [self["J4"][9 + index]]

    def get_driver_grounds(self, index):
        """Return the index'th driver ground pin for the single connector."""
        return [self["J3"][5 + index]]


class FastNC(FastNetBoard):

    """A FAST NANO Controller board."""

    def __init__(self):
        """Initialize the connectors on the NC."""
        super().__init__("NC")
        led_connector: PinSpecs = [("GND", 0), ("DO", 5), ("5v", 5)]
        self.add_connector("J1", led_connector)
        self.add_connector("J2", led_connector)
        self.add_connector("J4", led_connector)
        self.add_connector("J5", led_connector)
        self.add_connector("J7", [
            ("12v", 12),
            ("12v", 12),
            ("5v", 5),
            ("5v", 5),
            ("GND", 0),
            ("KEY", -1),
            ("GND", 0)
        ])
        self.add_connector("J10", self.fast_net_connector())
        self.add_connector("J11", self.fast_net_connector())

    def get_net_out(self):
        """Get the network output port, J10."""
        return self["J10"][0]

    def get_net_in(self):
        """Get the network input port, J11."""
        return self["J11"][0]


class FastSystem(System):

    """A complete FAST pinball based wiring system."""

    def __init__(self):
        """Initialize the system with basic contents."""
        super().__init__()
        # We at least will need a power board and nano controller
        self.pfb = FastPFB()
        self.nc = FastNC()
        self.add_board(self.pfb)
        self.add_board(self.nc)
        # Wire nano controller to power and ground on PFB
        self.connect(self.pfb["J4"][4], self.nc["J7"][2])
        self.connect(self.pfb["J4"][5], self.nc["J7"][3])
        self.connect(self.pfb["J4"][7], self.nc["J7"][0])
        self.connect(self.pfb["J4"][8], self.nc["J7"][4])
        self.connect(self.pfb["J4"][9], self.nc["J7"][6])


def light_numspec_to_ordinal(spec: str) -> int:
    """Turn a lighting specification into a light number.

    The MPF documentation says that FAST lights should be given by serial number only, but at least one
    example uses channel-index notation like for switches, and it seems to be accepted by the rest of MPF,
    so this allows for it.
    """
    if "-" in spec:
        parts = spec.split("-")
        return (int(parts[0]) * 64) + (int(parts[1]))
    return int(spec)


def wire_lights(machine: MachineController, s: FastSystem):
    """Wire lights for given machine definition into FAST system."""
    # Create a SerialLED board for each light on the system.
    led_boards: Dict[int, SerialLED] = dict()
    for light in machine.lights.values():
        new_board = SerialLED(light.name)
        real_num = light_numspec_to_ordinal(light.config["number"])
        if 0 <= real_num <= 255:
            led_boards[real_num] = new_board
            s.add_board(new_board)

    # List of bounds and names of FAST NC LED channels
    channels = [(0, 63, "J1"),
                (64, 127, "J2"),
                (128, 191, "J4"),
                (192, 255, "J5")]

    # Create a daisy chain for each light channel and connect it to the appropriate NC port.
    # TODO: These can be BIG - so big that dot can't render them in any format except SVG. Unfortunately
    #       wireviz will always try to render in all formats, giving an error. Laddering doesn't always help.
    for (start, end, ncport) in channels:
        channel_dict = {x: y for (x, y) in led_boards.items() if start <= x <= end}
        if len(channel_dict) > 0:
            # Using just one variable here will result in power wires not being connected at all if there is a
            # sequence gap, due to short circuit evaluation.
            ok1 = s.daisy_chain_dict(channel_dict, lambda x: x[""][0], lambda x: x[""][1], start, False)
            ok2 = s.daisy_chain_dict(channel_dict, lambda x: x[""][4], lambda x: x[""][4], start, False)
            ok3 = s.daisy_chain_dict(channel_dict, lambda x: x[""][5], lambda x: x[""][5], start, False)
            if not ok1 or not ok2 or not ok3:
                print("Lights numbers on channel", start, "-", end,
                      "aren't consecutive. Stopped at the first missing light.")
            s.connect(s.nc[ncport][1], channel_dict[start][""][1])
            s.connect(s.nc[ncport][2], channel_dict[start][""][4])
            s.connect(s.nc[ncport][0], channel_dict[start][""][5])


# pylint: disable=invalid-name
T = TypeVar("T")


def add_devices_with_board_numbers(devices: Dict[Any, T], s: FastSystem, construct: Callable[[T], Board])\
        -> Tuple[Dict[int, Dict[int, Board]], int]:
    """Build actual boards and board connection indices for all devices on a machine.

    Given a dict of devices from machine with config entries that specify board numbers, add boards for all
    the devices to the system and also build a dict associating each its board and port.
    :param devices The dict of devices. Only values are accessed.
    :param s The system to update.
    :param construct A function to build a Board from the dict values.
    :return A dictionary mapping boards numbers and devices to generated Boards, and the highest number board.
    """
    max_board: int = 0
    device_dict: Dict[int, Dict[int, Board]] = dict()
    for d in devices:
        sw = construct(d.name)
        s.add_board(sw)
        num_spec = d.config["number"]
        parts = num_spec.split("-")
        board_no = int(parts[0])
        if board_no > max_board:
            max_board = board_no
        device_no = int(parts[1])
        if board_no not in device_dict.keys():
            device_dict[board_no] = dict()
        device_dict[board_no][device_no] = sw
    return (device_dict, max_board)


def identify_fast_board(switches: int, drivers: int) -> Optional[FastIOBoard]:
    """Instantiate and return a FAST board capable of accommodating the given number of switches and drivers."""
    if switches > 32:
        return None
    if drivers > 16:
        return None
    if switches > 16:
        if drivers > 8:
            return None
        return FastIO3208()
    if switches <= 8:
        if drivers <= 4:
            return FastIO0804()
    return FastIO1616()


# pylint: disable=too-many-locals
def wire_specified_boards(machine: MachineController, s: FastSystem):
    """Wires a FAST system with specified board numbers."""
    # Used if switches and drivers specify board numbers.

    # Sort switches and drivers by board
    (fast_boards_switches, max_switch_board) = add_devices_with_board_numbers(machine.switches, s, Switch)
    (fast_boards_drivers, max_driver_board) = add_devices_with_board_numbers(machine.coils, s, Coil)

    # Check for board consistency
    max_board = max(max_switch_board, max_driver_board)
    for x in range(max_board + 1):
        if x not in fast_boards_switches and x not in fast_boards_drivers:
            print("I/O board numbers aren't consecutive. An empty board number", x, "would be needed!")
            return

    fast_boards: Dict[int, FastIOBoard] = dict()

    # Work out type of each board and add
    for board in range(max_board + 1):
        switches_on_board = len(fast_boards_switches[board])
        drivers_on_board = len(fast_boards_drivers[board])

        print("Board", board, "has", switches_on_board, "switches and", drivers_on_board, "drivers.")
        board_to_add = identify_fast_board(switches_on_board, drivers_on_board)
        if board_to_add is None:
            print(board, " has a combination of switches and drivers that no FAST IO board supports.")
            return
        s.add_board(board_to_add)
        fast_boards[board] = board_to_add

    # Build loop network
    s.daisy_chain_dict(fast_boards, lambda b: b.get_net_in(), lambda b: b.get_net_out(), 0, False)
    s.connect(s.nc.get_net_out(), fast_boards[0].get_net_in())
    s.connect(fast_boards[max_board].get_net_out(), s.nc.get_net_in())

    # Ground IO board logic connectors
    logic_grounds = [fast_boards[board].get_switch_grounds(0) for board in range(max_board + 1)]
    most_logic_slots = max([len(x) for x in logic_grounds])
    for slot in range(most_logic_slots):
        xth_slot_pins = ([g[slot] for g in logic_grounds if len(g) > slot])
        s.daisy_chain_list(xth_slot_pins, lambda p: p, lambda p: p)
        s.connect(s.pfb["J4"][8 + (slot % 2)], xth_slot_pins[0])

    # Ground HV connectors
    hv_grounds = [fast_boards[board].get_driver_grounds(0) for board in range(max_board + 1)]
    most_driver_slots = max([len(x) for x in hv_grounds])
    for slot in range(most_driver_slots):
        xth_slot_pins = ([g[slot] for g in hv_grounds if len(g) > slot])
        s.daisy_chain_list(xth_slot_pins, lambda p: p, lambda p: p)
        s.connect(s.pfb["J4"][2 + (slot % 2)], xth_slot_pins[0])

    # Wire up switches and drivers to board
    for board in range(max_board + 1):
        if board in fast_boards_switches.keys():
            for (pin, switch) in fast_boards_switches[board].items():
                s.connect(fast_boards[board].get_switch_pin(pin), switch[""][1])

        if board in fast_boards_drivers.keys():
            for (pin, driver) in fast_boards_drivers[board].items():
                s.connect(fast_boards[board].get_driver_pin(pin), driver[""][1])

    # Daisy chain switch and driver power
    for board in range(max_board + 1):
        if board in fast_boards_switches.keys():
            switches_on_board = list(fast_boards_switches[board].values())
            s.daisy_chain_list(switches_on_board, lambda w: w[""][0], lambda w: w[""][0])
            s.connect(s.pfb["J4"][4 + (board % 2)], switches_on_board[0][""][0])

        if board in fast_boards_drivers.keys():
            drivers_on_board = list(fast_boards_drivers[board].values())
            s.daisy_chain_list(drivers_on_board, lambda d: d[""][0], lambda d: d[""][0])
            s.connect(s.pfb["J4"][0 + (board % 2)], drivers_on_board[0][""][0])


def wire(machine: MachineController):
    """Wire a given machine using FAST Pinball boards."""
    s = FastSystem()
    wire_lights(machine, s)
    inconsistent_err = "Can't wire: all switch and driver numbers must be in the same format, (board-index) or " +\
                       "raw number, not a mixture of both."

    switches_specify_boards = None
    for switch in machine.switches.values():
        num_spec = switch.config["number"]
        if switches_specify_boards is None:
            switches_specify_boards = ("-" in num_spec)
        else:
            if ("-" in num_spec) != switches_specify_boards:
                print(inconsistent_err)
                return None

    drivers_specify_boards = None
    for coil in machine.coils.values():
        num_spec = coil.config["number"]
        if drivers_specify_boards is None:
            drivers_specify_boards = ("-" in num_spec)
        else:
            if ("-" in num_spec) != drivers_specify_boards:
                print(inconsistent_err)
                return None

    if switches_specify_boards != drivers_specify_boards:
        print(inconsistent_err)
        return None

    if switches_specify_boards:
        wire_specified_boards(machine, s)
    else:
        # TODO: the problem here is that we need to either guess, or work out from hardware, what types
        #       of boards are used.
        print("Can't currently wire FAST pinball if board numbers are not specified.")

    return s.dump()



from mpf.devices.light import Light

from mpf.tests.MpfTestCase import MpfTestCase


class TestDeviceMatrixLight(MpfTestCase):

    def get_config_file(self):
        return 'matrix_lights.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light/'

    def _synchronise_light_update(self):
        ts = Light._updater_task.get_next_call_time()
        self.assertTrue(ts)
        self.advance_time_and_run(ts - self.machine.clock.get_time())
        self.advance_time_and_run(0.01)

    def testBasicOnAndOff(self):
        """Tests setting some light brightness levels (using default arguments)"""

        # Make sure hardware devices have been configured for tests
        self.assertIn('light_01', self.machine.lights)
        self.assertIn('light_02', self.machine.lights)

        light1 = self.machine.lights["light_01"]
        light2 = self.machine.lights["light_02"]

        self.advance_time_and_run(10)

        # Lights should start out off (current brightness is 0)
        self.assertLightChannel("light_01", 0)
        self.assertLightChannel("light_02", 0)

        light1.on(128)
        self.advance_time_and_run(1)
        self.assertEqual([128, 128, 128], light1.get_color())
        self.assertAlmostEqual(self.machine.clock.get_time() - 1,
                               light1.stack[0].start_time)
        self.assertLightChannel("light_01", 128)
        self.assertEqual(0, light1.stack[0].priority)

        light2.on(255)
        self.advance_time_and_run(1)
        self.assertLightChannel("light_02", 255)
        self.assertEqual([255, 255, 255], light2.get_color())
        self.assertAlmostEqual(self.machine.clock.get_time() - 1,
                               light2.stack[0].start_time)
        self.assertEqual(0, light2.stack[0].priority)

        # Turn the lights off
        light1.off()
        self.advance_time_and_run(1)
        self.assertEqual([0, 0, 0], light1.get_color())
        self.assertAlmostEqual(self.machine.clock.get_time() - 1,
                               light1.stack[0].start_time)
        self.assertLightChannel("light_01", 0)
        self.assertEqual(0, light1.stack[0].priority)

        light2.off()
        self.advance_time_and_run(1)
        self.assertEqual([0, 0, 0], light2.get_color())
        self.assertAlmostEqual(self.machine.clock.get_time() - 1,
                               light2.stack[0].start_time)
        self.assertLightChannel("light_02", 0)
        self.assertEqual(0, light2.stack[0].priority)



"""An achievement which can be reached in a pinball machine."""

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler

from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player


@DeviceMonitor("state", "selected")
class Achievement(ModeDevice):

    """An achievement in a pinball machine.

    It is tracked per player and can automatically restore state on the next
    ball.
    """

    config_section = 'achievements'
    collection = 'achievements'
    class_label = 'achievement'
    allow_empty_configs = True

    def __init__(self, machine, name):
        """Initialise achievement."""
        super().__init__(machine, name)
        self._player = None
        self._mode = None
        self._show = None

    @property
    def state(self):
        """Return current state."""
        try:
            return self._player.achievements[self.name][0]
        except (AttributeError, KeyError):
            return None

    @state.setter
    def state(self, value):
        """Set current state."""
        try:
            self._player.achievements[self.name][0] = value
        except (AttributeError, KeyError):
            self._player.achievements[self.name] = [value, False]

    @property
    def can_be_selected_for_start(self):
        """Return if this achievement can be selected and started."""
        state = self.state
        return state == 'enabled' or (state == 'stopped' and self.config['restart_after_stop_possible'])

    @property
    def selected(self):
        """Return current selection state."""
        try:
            return self._player.achievements[self.name][1]
        except (AttributeError, KeyError):
            return False

    @selected.setter
    def selected(self, value):
        """Set current selected."""
        try:
            self._player.achievements[self.name][1] = value
        except (AttributeError, KeyError):
            self._player.achievements[self.name] = [None, value]

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Validate and parse config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)

        states = ['disabled', 'enabled', 'started', 'stopped', 'selected', 'completed']

        for state in states:
            if not config['events_when_{}'.format(state)]:
                config['events_when_{}'.format(state)] = [
                    "achievement_{}_state_{}".format(self.name, state)]

        return config

    def enable(self):
        """Enable the achievement.

        It can only start if it was enabled before.
        """
        super().enable()
        if self.state in ("disabled", "started"):
            self.state = "enabled"
            self._run_state()

    @event_handler(5)
    def event_start(self, **kwargs):
        """Event handler for start event."""
        del kwargs
        self.start()

    def start(self):
        """Start achievement."""
        if self.state == "enabled" or (
            self.config['restart_after_stop_possible'] and
                self.state == "stopped"):
            self.state = "started"
            self.selected = False
            self._run_state()

    @event_handler(4)
    def event_complete(self, **kwargs):
        """Event handler for complete event."""
        del kwargs
        self.complete()

    def complete(self):
        """Complete achievement."""
        if self.state == "started":
            self.state = "completed"
            self.selected = False
            self._run_state()

    @event_handler(2)
    def event_stop(self, **kwargs):
        """Event handler for stop event."""
        del kwargs
        self.stop()

    def stop(self):
        """Stop achievement."""
        if self.state == "started":
            self.state = "stopped"
            self.selected = False
            self._run_state()

    @event_handler(0)
    def event_disable(self, **kwargs):
        """Event handler for disable event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable achievement."""
        if self.state == "enabled" or (
            self.config['restart_after_stop_possible'] and
                self.state == "stopped"):
            self.state = "disabled"
            self.selected = False
            self._run_state()

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset the achievement to its initial state."""
        # if there is no player active
        if not self._player:
            return

        self.selected = False

        if self.config['start_enabled'] is True:
            self.state = "enabled"
        elif self.config['start_enabled'] is False:
            self.state = "disabled"
        elif self.config['enable_events']:
            self.state = "disabled"
        else:
            self.state = "enabled"

        self._run_state()

    @event_handler(8)
    def event_unselect(self, **kwargs):
        """Event handler for unselect event."""
        del kwargs
        self.unselect()

    def unselect(self):
        """Remove highlight (unselect) this achievement."""
        if not self._player:
            return

        self.debug_log("Unselecting achievement")

        if self.selected:
            self.selected = False

            self._run_state()

    @event_handler(9)
    def event_select(self, **kwargs):
        """Event handler for select event."""
        del kwargs
        self.select()

    def select(self):
        """Highlight (select) this achievement."""
        if not self._player:
            return

        self.debug_log("Selecting achievement")

        if (self.state == 'enabled' or
                (self.config['restart_after_stop_possible'] and self.state == "stopped")) and not self.selected:
            self.selected = True
            self._run_state()

    def _run_state(self, restore=False):
        """Run shows and post events for current step."""
        self.machine.events.post("achievement_{}_changed_state".format(self.name), restore=restore,
                                 state=self.state, selected=self.selected)
        '''event: achievement_(name)_changed_state
        desc: Achievement (name) changed state.

        Valid states are: disabled, enabled, started, completed, stopped

        This is only posted once per state. Its also posted on restart on the next ball to restore state.

        args:
            restore: true if this is reposted to restore state
            state: Current state
            selected: Whatever this achievement is selected currently

        '''
        for event in self.config['events_when_{}'.format(self.state)]:
            self.machine.events.post(event, restore=restore, state=self.state, selected=self.selected)
            '''event: achievement_(name)_state_(state)
            desc: Achievement (name) changed to state (state).

            Valid states are: disabled, enabled, started, completed, stopped

            This is only posted once per state. Its also posted on restart on the next ball to restore state
            and when selection changes.

            args:
                restore: true if this is reposted to restore state
                state: Current state
                selected: Whatever this achievement is selected currently

            '''
        if self.selected:
            for event in self.config['events_when_selected']:
                self.machine.events.post(event, restore=restore, state=self.state, selected=self.selected)
                # same as above

        if self._show:
            self.debug_log('Stopping show: %s', self._show)
            self._show.stop()
            self._show = None

        if self.selected and self.config['show_when_selected']:
            show = self.config['show_when_selected']
        else:
            show = self.config['show_when_' + self.state]

        if show:
            self.debug_log('Playing show: %s. Priority: %s. Loops: -1. '
                           'Show tokens: %s', show.name, self._mode.priority,
                           self.config['show_tokens'])

            self._show = show.play(
                priority=self._mode.priority,
                loops=-1, sync_ms=self.config['sync_ms'],
                show_tokens=self.config['show_tokens'])

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Load device on mode start and restore state.

        Args:
        ----
            mode: mode which was contains the device
            player: player which is currently active
        """
        self._player = player
        self._mode = mode
        if not self._player.achievements:
            self._player.achievements = dict()      # type: ignore

        if self.name not in self._player.achievements:
            self.reset()
        else:
            self._restore_state()

        # state might have changed
        self.notify_virtual_change("selected", None, self.state)    # type: ignore

    def _restore_state(self):
        if self.state == "started" and not (
                self.config['restart_on_next_ball_when_started']):
            self.state = "stopped"
        elif self.state == "enabled" and not (
                self.config['enable_on_next_ball_when_enabled']):
            self.state = "disabled"
        else:
            # state might still have changed because of player change
            self.notify_virtual_change("state", None, self.state)

        self._run_state(restore=True)

    def device_removed_from_mode(self, mode: Mode):
        """Mode ended.

        Args:
        ----
            mode: mode which stopped
        """
        del mode
        self._player = None
        self._mode = None
        if self._show:
            self._show.stop()
            self._show = None

    def add_control_events_in_mode(self, mode: Mode) -> None:
        """Override the default mode device behavior.

        Achievements use sophisticated logic to handle their mode-starting states
        during device_loaded_in_mode(). Therefore no default enabling is required.
        """



from unittest.mock import MagicMock

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestScoreQueue(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/score_queue/'

    def test_score_queue(self):
        """Test the score queue for SS games."""
        # try a score outside of a game. should not crash
        self.machine.score_queues["score"].score(2000)
        self.advance_time_and_run()

        # start game with two players
        self.start_two_player_game()
        self.machine.coils["c_chime_1000"].pulse = MagicMock()
        self.machine.coils["c_chime_100"].pulse = MagicMock()
        self.machine.coils["c_chime_10"].pulse = MagicMock()

        self.assertPlayerVarEqual(0, "score")
        self.machine.score_queues["score"].score(2100)
        self.machine.score_queues["score"].score(20)
        # first queue entry
        self.advance_time_and_run(.1)
        self.assertPlayerVarEqual(1000, "score")
        self.machine.coils["c_chime_1000"].pulse.assert_called_with()
        self.machine.coils["c_chime_1000"].pulse = MagicMock()
        self.advance_time_and_run(.2)
        self.assertPlayerVarEqual(2000, "score")
        self.machine.coils["c_chime_1000"].pulse.assert_called_with()
        self.machine.coils["c_chime_1000"].pulse = MagicMock()
        self.machine.coils["c_chime_100"].pulse.assert_not_called()
        self.advance_time_and_run(.2)
        self.assertPlayerVarEqual(2100, "score")
        self.machine.coils["c_chime_100"].pulse.assert_called_with()
        self.machine.coils["c_chime_100"].pulse = MagicMock()
        self.machine.coils["c_chime_1000"].pulse.assert_not_called()
        self.machine.coils["c_chime_10"].pulse.assert_not_called()
        self.advance_time_and_run(.2)
        # second queue entry now
        self.assertPlayerVarEqual(2110, "score")
        self.machine.coils["c_chime_10"].pulse.assert_called_with()
        self.machine.coils["c_chime_10"].pulse = MagicMock()
        self.machine.coils["c_chime_1000"].pulse.assert_not_called()
        self.machine.coils["c_chime_100"].pulse.assert_not_called()
        self.advance_time_and_run(.2)
        self.assertPlayerVarEqual(2120, "score")
        self.machine.coils["c_chime_10"].pulse.assert_called_with()
        self.machine.coils["c_chime_10"].pulse = MagicMock()
        self.machine.coils["c_chime_1000"].pulse.assert_not_called()
        self.machine.coils["c_chime_100"].pulse.assert_not_called()

        # no more
        self.advance_time_and_run(1)
        self.assertPlayerVarEqual(2120, "score")
        self.machine.coils["c_chime_1000"].pulse.assert_not_called()
        self.machine.coils["c_chime_100"].pulse.assert_not_called()
        self.machine.coils["c_chime_10"].pulse.assert_not_called()

        # test scoring during drain
        self.machine.score_queues["score"].score(90)
        self.drain_all_balls()

        self.advance_time_and_run(2)
        self.assertEqual(2210, self.machine.game.player_list[0].score)
        self.assertEqual(0, self.machine.game.player_list[1].score)

        self.assertPlayerNumber(2)
        self.machine.score_queues["score"].score(90)
        self.advance_time_and_run(2)
        self.assertEqual(2210, self.machine.game.player_list[0].score)
        self.assertEqual(90, self.machine.game.player_list[1].score)

        self.start_mode("mode1")
        self.post_event("score_2k")
        self.post_event("score_200")
        self.advance_time_and_run(2)
        self.assertEqual(2210, self.machine.game.player_list[0].score)
        self.assertEqual(2290, self.machine.game.player_list[1].score)



"""Base class for asyncio modes."""
from typing import Optional

import abc
import asyncio

from mpf.core.mode import Mode

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class AsyncMode(Mode, metaclass=abc.ABCMeta):

    """Base class for asyncio modes."""

    __slots__ = ["_task"]

    def __init__(self, machine: "MachineController", *args, **kwargs) -> None:
        """Initialise async mode."""
        super().__init__(machine, *args, **kwargs)

        self._task = None   # type: Optional[asyncio.Task]

    def _started(self, **kwargs) -> None:
        """Start main task."""
        del kwargs
        super()._started()

        self._task = self.machine.clock.loop.create_task(self._run())
        self._task.add_done_callback(self._mode_ended)

    def _mode_ended(self, future: asyncio.Future) -> None:
        """Evaluate result of task.

        Will raise exceptions from within task.
        """
        try:
            future.result()
        except asyncio.CancelledError:
            pass

        # stop mode
        self.stop()

    def _stopped(self) -> None:
        """Cancel task."""
        super()._stopped()

        if self._task:
            self._task.cancel()
            self._task = None

    @abc.abstractmethod
    async def _run(self) -> None:
        """Start main task which runs as long as the mode is active.

        Overwrite this function in your mode.

        Its automatically canceled when the mode stops. You can catch CancelError to handle mode stop.
        """



"""Routes balls from one device to another when captured."""
from collections import defaultdict
from typing import List, Dict

from mpf.core.device_monitor import DeviceMonitor

from mpf.core.enable_disable_mixin import EnableDisableMixin
from mpf.core.events import EventHandlerKey

from mpf.devices.ball_device.ball_device import BallDevice

from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController   # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("balls_routing")
class BallRouting(EnableDisableMixin, ModeDevice):

    """Route balls from one device to another when captured."""

    config_section = 'ball_routings'
    collection = 'ball_routings'
    class_label = 'ball_routing'

    __slots__ = ["_routing_queue", "_balls_at_target", "_handler"]

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise device."""
        super().__init__(machine, name)
        self._routing_queue = defaultdict(int)      # type: Dict[BallDevice, int]
        self._balls_at_target = 0
        self._handler = []                          # type: List[EventHandlerKey]

    def _enable(self):
        """Enable routing."""
        for device in self.config['source_devices']:
            self._handler.append(
                self.machine.events.add_handler(
                    'balldevice_' + device.name + '_ball_enter',
                    self._claim_balls, device=device, priority=self.mode.priority))
            self._handler.append(
                self.machine.events.add_handler(
                    'balldevice_' + device.name + '_ball_entered',
                    self._route_ball, device=device, priority=self.mode.priority))

        self._handler.append(
            self.machine.events.add_handler(
                'balldevice_' + self.config['target_device'].name + '_ball_enter',
                self._add_ball, priority=self.mode.priority + 10000))
        self.debug_log("Enabling")

    def _claim_balls(self, device: BallDevice, unclaimed_balls, **kwargs):
        """Claim balls to route them to destination later."""
        del kwargs
        if not self._enabled:
            return {}
        # remember how many balls were captured
        self._routing_queue[device] += unclaimed_balls
        # claim all balls
        return {"unclaimed_balls": 0}

    def _route_ball(self, device: BallDevice, **kwargs):
        """Route balls to destination."""
        del kwargs
        self._balls_at_target += self._routing_queue[device]
        for _ in range(self._routing_queue[device]):
            device.eject(target=self.config['target_device'])

        self._routing_queue[device] = 0

    def _add_ball(self, unclaimed_balls, new_balls, **kwargs):
        """Mark balls as unclaimed at destination."""
        del kwargs
        claimed_balls = new_balls - unclaimed_balls
        if self._balls_at_target and claimed_balls:
            if claimed_balls <= self._balls_at_target:
                new_unexpected = unclaimed_balls + self._balls_at_target
                self.log.debug("Adding %s balls to target %s", new_unexpected, self.config['target_device'].name)
                self._balls_at_target = 0
                return {"unclaimed_balls": new_unexpected}

            self.log.debug("Adding %s balls to target %s", claimed_balls, self.config['target_device'].name)
            self._balls_at_target -= claimed_balls
            return {"unclaimed_balls": new_balls}

        return {}

    def _disable(self):
        """Disable routing."""
        self.debug_log("Disabling")
        self._enabled = False
        self.machine.events.remove_handlers_by_keys(self._handler)
        self._handler = []

    def device_removed_from_mode(self, mode: Mode) -> None:
        """Disable ball save when mode ends."""
        super().device_removed_from_mode(mode)
        self.debug_log("Removing")
        self.disable()



from unittest.mock import MagicMock, call

from mpf.config_players.plugin_player import PluginPlayer
from mpf.tests.MpfBcpTestCase import MpfBcpTestCase
from mpf.tests.MpfTestCase import TestMachineController


# Override the plugin player functionality so that it pulls in our test one
def _register_plugin_config_players(self):
    TestConfigPlayer.register_with_mpf(self)
    TestConfigPlayer2.register_with_mpf(self)
    TestConfigPlayer3.register_with_mpf(self)

TestMachineController._register_plugin_config_players = _register_plugin_config_players     # type: ignore


class TestConfigPlayer(PluginPlayer):
    config_file_section = 'test_player'
    show_section = 'tests'

    def get_express_config(self, value):
        return dict(some=value)

    def validate_config(self, config):
        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event][self.show_section] = dict()

            if not isinstance(settings, dict):
                settings = {settings: dict()}

        return validated_config

    def _validate_config_item(self, device, device_settings):
        if device_settings is None:
            device_settings = device

        if not isinstance(device_settings, dict):
            device_settings = self.get_express_config(device_settings)

        devices = [device]

        return_dict = dict()
        for device in devices:
            return_dict[device] = device_settings

        return return_dict

    @staticmethod
    def register_with_mpf(machine):
        return 'test', TestConfigPlayer(machine)


class TestConfigPlayer2(PluginPlayer):
    config_file_section = 'test2_player'
    show_section = 'test2s'

    def get_express_config(self, value):
        return dict(some=value)

    def validate_config(self, config):
        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event][self.show_section] = dict()
            if not isinstance(settings, dict):
                settings = {settings: dict()}

        return validated_config

    def _validate_config_item(self, device, device_settings):
        if device_settings is None:
            device_settings = device

        if not isinstance(device_settings, dict):
            device_settings = self.get_express_config(device_settings)

        devices = [device]

        return_dict = dict()
        for device in devices:
            return_dict[device] = device_settings

        return return_dict

    @staticmethod
    def register_with_mpf(machine):
        return 'test2', TestConfigPlayer2(machine)


class TestConfigPlayer3(PluginPlayer):
    config_file_section = 'test3_player'
    show_section = 'test3s'

    def get_express_config(self, value):
        return dict(some=value)

    def validate_config(self, config):
        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event][self.show_section] = dict()
            if not isinstance(settings, dict):
                settings = {settings: dict()}

        return validated_config

    def _validate_config_item(self, device, device_settings):
        if device_settings is None:
            device_settings = device

        if not isinstance(device_settings, dict):
            device_settings = self.get_express_config(device_settings)

        devices = [device]

        return_dict = dict()
        for device in devices:
            return_dict[device] = device_settings

        return return_dict

    @staticmethod
    def register_with_mpf(machine):
        return 'test3', TestConfigPlayer3(machine)


class TestPluginConfigPlayer(MpfBcpTestCase):
    def get_config_file(self):
        return 'plugin_config_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/plugin_config_player/'

    def setUp(self):
        self.machine_spec_patches['test_player'] = dict(__valid_in__='machine, mode')
        self.machine_spec_patches['test2_player'] = dict(__valid_in__='machine, mode')

        super().setUp()
        self._bcp_client.send = MagicMock()

    def test_plugin_config_player(self):
        # Setup BCP to monitor mode events
        self._bcp_client.receive_queue.put_nowait(('monitor_start', {'category': 'modes'}))
        self.advance_time_and_run()

        self.assertIn('tests', self.machine.show_controller.show_players)
        self.assertIn('test2s', self.machine.show_controller.show_players)

        # event1 is in the test_player only. Check that it's sent as a
        # trigger
        self._bcp_client.send.reset_mock()
        self.machine.events.post('event1')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'priority': 0, 'settings': {'tests': {}},
                                                             'name': 'tests_play', 'calling_context': 'event1',
                                                             'context': '_global'})

        self.assertEqual(1, len(self.machine.bcp.transport._handlers["tests_play"]))
        self._bcp_client.send.reset_mock()
        self.machine.events.post('event1')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'priority': 0, 'settings': {'tests': {}},
                                                             'name': 'tests_play', 'calling_context': 'event1',
                                                             'context': '_global'})

        self.assertEqual(1, len(self.machine.bcp.transport._handlers["tests_play"]))
        self._bcp_client.send.reset_mock()

        # event2 is in the test_player and test2_player. Check that it's only
        # sent once
        self.machine.events.post('event2')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'calling_context': 'event2', 'context': '_global',
                                                             'priority': 0, 'name': 'test2s_play',
                                                             'settings': {'test2s': {}}})
        self._bcp_client.send.reset_mock()

        # event3 is test2_player only. Check that it's only sent once
        self.machine.events.post('event3')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'priority': 0, 'name': 'test2s_play',
                                                             'context': '_global', 'settings': {'test2s': {}},
                                                             'calling_context': 'event3'})
        self._bcp_client.send.reset_mock()

        # fake_event isn't used in any player. Check that it's not sent
        self.machine.events.post('fake_event')
        self.advance_time_and_run()
        assert not self._bcp_client.send.called

        # Start mode1
        self.machine.modes['mode1'].start()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes['mode1'].active)
        self._bcp_client.send.assert_called_with
        self._bcp_client.send.assert_has_calls([
            call('mode_start', {'name': 'mode1', 'priority': 400}),
            call('mode_list', {'running_modes': [('mode1', 400), ('attract', 10)]})
        ])
        self._bcp_client.send.reset_mock()

        # event4 is in test_player for mode1, so make sure it sends now
        self.machine.events.post('event4')
        self.advance_time_and_run()

        self._bcp_client.send.assert_called_with('trigger', {'settings': {'tests': {}}, 'calling_context': 'event4',
                                                             'priority': 400, 'context': 'mode1', 'name': 'tests_play'})
        self._bcp_client.send.reset_mock()

        # Stop mode 1
        self.machine.modes['mode1'].stop()
        self.advance_time_and_run()
        self._bcp_client.send.assert_has_calls([
            call('trigger', {'context': 'mode1', 'name': 'tests_clear'}),
            call('trigger', {'context': 'mode1', 'name': 'test2s_clear'}),
            call('mode_stop', {'name': 'mode1'}),
            call('mode_list', {'running_modes': [('attract', 10)]})
        ])
        self._bcp_client.send.reset_mock()

        # post event4 again, and it should not be sent since that mode was
        # stopped
        self.machine.events.post('event4')
        self.advance_time_and_run()
        assert not self._bcp_client.send.called

        # event1, event2, and event3 should still work. Even though they were
        # in mode1, they were also in the base config

        # event1
        self.machine.events.post('event1')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'settings': {'tests': {}}, 'calling_context': 'event1',
                                                             'name': 'tests_play', 'context': '_global', 'priority': 0})
        self._bcp_client.send.reset_mock()

        # event2
        self.machine.events.post('event2')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'settings': {'test2s': {}}, 'context': '_global',
                                                             'calling_context': 'event2', 'name': 'test2s_play',
                                                             'priority': 0})
        self._bcp_client.send.reset_mock()

        # event3
        self.machine.events.post('event3')
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_with('trigger', {'context': '_global', 'name': 'test2s_play',
                                                             'calling_context': 'event3', 'settings': {'test2s': {}},
                                                             'priority': 0})
        self._bcp_client.send.reset_mock()

    def test_plugin_from_show(self):
        t1_player = self.machine.show_controller.show_players['tests']
        t1_player.play = MagicMock()

        self.machine.shows['show1'].play()
        self.advance_time_and_run()

        self.assertTrue(t1_player.play.called)
        self.assertEqual(1, len(self.machine.bcp.transport._handlers["tests_play"]))
        self.assertEqual(1, len(self.machine.bcp.transport._handlers["tests_clear"]))

        self.assertIn('tests_play', self.machine.bcp.transport._handlers)
        self.assertIn('tests_clear', self.machine.bcp.transport._handlers)
        self.assertIn('test2s_play', self.machine.bcp.transport._handlers)
        self.assertIn('test2s_clear', self.machine.bcp.transport._handlers)

    def test_conditional_events(self):
        self.machine.events.post('event5')
        self.advance_time_and_run()
        self._bcp_client.send.assert_not_called()
        self._bcp_client.send.reset_mock()

        self.machine.events.post('event5', foo=0)
        self.advance_time_and_run()
        self._bcp_client.send.assert_called_once_with('trigger', {'settings': {'tests': {}}, 'name': 'tests_play',
                                                                  'priority': 0, 'context': '_global',
                                                                  'calling_context': 'event5{foo==0}',
                                                                  'foo': 0})
        self._bcp_client.send.reset_mock()

    def test_plugin_in_show_but_not_standalone(self):
        # tests a plugin player that has an entry in a show, but that does
        # not have a config player entry section in any other config file

        # e.g. a "slides" section in a show, but now "slide_player:"
        # section anywhere

        t3_player = self.machine.show_controller.show_players['test3s']
        t3_player.play = MagicMock()

        self.machine.modes['mode1'].start()
        self.advance_time_and_run()

        self.post_event('start_show3')

        self.assertIn('test3s_play', self.machine.bcp.transport._handlers)
        self.assertIn('test3s_clear', self.machine.bcp.transport._handlers)

        self.assertTrue(t3_player.play.called)



"""Pololu TIC controller platform."""
import asyncio
import logging

from mpf.exceptions.config_file_error import ConfigFileError
from mpf.platforms.pololu.pololu_ticcmd_wrapper import PololuTiccmdWrapper
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface
from mpf.core.platform import StepperPlatform


class PololuTICHardwarePlatform(StepperPlatform):

    """Supports the Pololu TIC stepper drivers via ticcmd command line."""

    def __init__(self, machine):
        """Initialise TIC platform."""
        super().__init__(machine)
        self.config = self.machine.config_validator.validate_config("pololu_tic",
                                                                    self.machine.config.get('pololu_tic', {}))
        self._configure_device_logging_and_debug("Pololu TIC", self.config)
        self.features['tickless'] = True
        self._steppers = []

    def __repr__(self):
        """Return string representation."""
        return '<Platform.PololuTICHardwarePlatform>'

    def stop(self):
        """De-energize the stepper and stop sending the command timeout refresh."""
        for stepper in self._steppers:
            stepper.shutdown()
        self._steppers = []

    async def configure_stepper(self, number: str, config: dict) -> "PololuTICStepper":
        """Configure a smart stepper device in platform.

        Args:
        ----
            number: Number of this stepper.
            config (dict): Configuration of device
        """
        stepper = PololuTICStepper(number, config, self.machine)
        self._steppers.append(stepper)
        await stepper.initialize()
        return stepper

    @classmethod
    def get_stepper_config_section(cls):
        """Return config validator name."""
        return "tic_stepper_settings"


class PololuTICStepper(StepperPlatformInterface):

    """A stepper on a pololu TIC."""

    def __init__(self, number, config, machine):
        """Initialise stepper."""
        self.config = config
        self.log = logging.getLogger('TIC Stepper')
        self.log.debug("Configuring Stepper Parameters.")
        self.serial_number = number
        self.tic = PololuTiccmdWrapper(self.serial_number, machine, False)
        self.machine = machine
        self._position = None
        self._watchdog_task = None

        if self.config['step_mode'] not in [1, 2, 4, 8, 16, 32]:
            raise ConfigFileError("step_mode must be one of (1, 2, 4, 8, 16, or 32)", 1, self.log.name)

        if self.config['max_speed'] <= 0:
            raise ConfigFileError("max_speed must be greater than 0", 2, self.log.name)

        if self.config['max_speed'] > 500000000:
            raise ConfigFileError("max_speed must be less than or equal to 500,000,000", 3, self.log.name)

    async def initialize(self):
        """Configure the stepper."""
        self.log.debug("Looking for TIC Device with serial number %s.", self.serial_number)
        status = await self.tic.get_status()

        if "Low VIN" in status['Errors currently stopping the motor']:
            self.log.debug("WARNING: Reporting Low VIN")

        self._position = status['Current position']

        self.log.debug("TIC Status: ")
        self.log.debug(status)

        self.tic.set_step_mode(self.config['step_mode'])
        self.tic.set_max_speed(self.config['max_speed'])
        self.tic.set_starting_speed(self.config['starting_speed'])
        self.tic.set_max_acceleration(self.config['max_acceleration'])
        self.tic.set_max_deceleration(self.config['max_deceleration'])
        self.tic.set_current_limit(self.config['current_limit'])

        self.tic.exit_safe_start()
        self.tic.energize()

        self._watchdog_task = self.machine.clock.schedule_interval(self._reset_command_timeout, .5)

    def _reset_command_timeout(self):
        """Reset the command timeout."""
        self.tic.reset_command_timeout()

    def home(self, direction):
        """Home an axis, resetting 0 position."""
        self.tic.halt_and_hold()    # stop the stepper in case its moving
        self._position = 0
        if direction == 'clockwise':
            self.log.debug("Homing clockwise")
            self.tic.go_home(True)
        elif direction == 'counterclockwise':
            self.log.debug("Homing counterclockwise")
            self.tic.go_home(False)

    def move_abs_pos(self, position):
        """Move axis to a certain absolute position."""
        self.log.debug("Moving To Absolute Position: %s", position)
        self._position = position
        self.tic.rotate_to_position(self._position)

    def move_rel_pos(self, position):
        """Move axis to a relative position."""
        self._position += position
        self.log.debug("Moving To Relative Position: %s (Absolute: %s)", position, self._position)
        self.tic.rotate_to_position(self._position)

    def move_vel_mode(self, velocity):
        """Move at a specific velocity and direction (pos = clockwise, neg = counterclockwise)."""
        if velocity == 0:
            self.log.debug("Stopping Due To Velocity Set To 0")
            self.tic.halt_and_hold()     # motor stop
        else:
            calculated_velocity = velocity * self.config['max_speed']
            self.log.debug("Rotating By Velocity (velocity * max_speed): %s", calculated_velocity)
            self.tic.rotate_by_velocity(calculated_velocity)

    def set_home_position(self):
        """Set position to home."""
        self.set_position(0)

    def set_position(self, position):
        """Set the current position of the stepper.

        Args:
        ----
            position (number): The position to set
        """
        self.log.debug("Setting Position To %s", position)
        self._position = position
        self.tic.halt_and_set_position(position)

    def stop(self) -> None:
        """Stop stepper."""
        self.log.debug("Commanded To Stop")
        self.tic.halt_and_hold()

    def shutdown(self):
        """Shutdown stepper."""
        if self._watchdog_task:
            self._watchdog_task.cancel()
            self._watchdog_task = None
        self.tic.halt_and_hold()
        self.tic.stop()

    async def wait_for_move_completed(self):
        """Wait until move completed."""
        while not await self.is_move_complete():
            await asyncio.sleep(1 / self.config['poll_ms'])

    async def is_move_complete(self) -> bool:
        """Return true if move is complete."""
        status = await self.tic.get_status()
        current_position = status['Current position']
        self.log.debug("Target Position: %s Current Position: %s", self._position, current_position)
        return bool(self._position == current_position)



"""Migrates YAML configuration files for MPF from one version to another."""

import os
import shutil
import time
import datetime
import logging
import importlib
from copy import deepcopy

from ruamel import yaml
from ruamel.yaml.comments import CommentedSeq, CommentedMap

from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip
from mpf._version import version
from mpf.core.file_manager import FileManager
from mpf.core.utility_functions import Util

EXTENSION = '.yaml'
BACKUP_FOLDER_NAME = 'previous_config_files'
TARGET_CONFIG_VERSION = 4  # todo change to dynamic param
REPROCESS_CURRENT_VERSION = False
INDENTATION_SPACES = 4


class Migrator:

    """Migrates a config."""

    def __init__(self, mpf_path, machine_path):
        """Initialise migrator."""
        self.log = logging.getLogger('Migrator')
        self.start_time = time.time()
        self.num_config_files = 0
        self.num_show_files = 0
        self.log.info("MPF Migrator: %s", version)
        self.log.info("Migrating config and show files from: %s",
                      machine_path)

        self.mpf_path = mpf_path
        self.machine_path = machine_path
        self.file_list = list()
        self.backup_folder = None
        self.base_folder = None
        # self.target_config_version = int(__config_version__)
        self.target_config_version = TARGET_CONFIG_VERSION
        self.log.info("New config version will be v%s",
                      self.target_config_version)

        self.migrator = importlib.import_module(
            'mpf.migrator.config_version_{}'.format(TARGET_CONFIG_VERSION))

        self.log.debug("Found Migrator for config files v%s",
                       TARGET_CONFIG_VERSION)

        self.build_file_list()
        self.backup_files()
        self.migrate_files()

    def build_file_list(self):
        """Build file list for machine."""
        for root, _, files in os.walk(self.machine_path):
            for file in files:
                if (os.path.splitext(file)[1].lower() == EXTENSION and
                        BACKUP_FOLDER_NAME not in root):
                    self.log.debug("Found file: %s", os.path.join(root, file))
                    self.file_list.append(os.path.join(root, file))

        self.base_folder = (os.path.commonprefix(self.file_list))
        self.log.debug("Detected base folder: %s", self.base_folder)

    def create_backup_folder(self):
        """Create backup folder."""
        this_time = datetime.datetime.fromtimestamp(time.time()).strftime(
            '%Y-%m-%d-%H-%M-%S')
        self.backup_folder = os.path.join(self.machine_path,
                                          BACKUP_FOLDER_NAME, this_time)

    def backup_files(self):
        """Create folder and backup config."""
        self.create_backup_folder()
        self.log.info("Backing up files to %s", self.backup_folder)
        shutil.copytree(self.base_folder, self.backup_folder,
                        ignore=self._ignore_files)
        self.log.debug("Backup done")

    @classmethod
    def _ignore_files(cls, root_folder, contents):
        # ignore previous backup folders
        if BACKUP_FOLDER_NAME in root_folder:
            return contents

        # ignore the data folder since those files can't be opened
        # with the round trip loaded
        if root_folder.split(os.sep)[-1] == 'data':
            return contents

        return_list = list()

        # ignore everything that doesn't have the target extension
        for item in contents:
            if (os.path.isfile(os.path.join(root_folder, item)) and
                    os.path.splitext(item)[1].lower() != EXTENSION):
                return_list.append(item)

        return return_list

    def migrate_files(self):
        """Migrate files in machine."""
        # We need to migrate show files after config files since we need
        # to pull some things out of the configs for the shows.
        round2_files = list()

        for file in self.file_list:
            file_content = FileManager.load(file)

            if isinstance(file_content, CommentedMap):
                migrated_content = self.migrator.migrate_file(file,
                                                              file_content)
                if migrated_content:
                    self.num_config_files += 1
                    self.save_file(file, migrated_content)
            else:
                round2_files.append(file)

        for file in round2_files:
            file_content = FileManager.load(file)
            migrated_content = self.migrator.migrate_file(file, file_content)
            if migrated_content:
                self.num_show_files += 1
                self.save_file(file, migrated_content)

        self.log.info("DONE! Migrated %s config file(s) and %s show file(s) in"
                      " %ss", self.num_config_files, self.num_show_files,
                      round(time.time() - self.start_time, 2))
        self.log.info("Detailed log file is in %s.", os.path.join(
            self.machine_path, 'logs'))
        self.log.info("Original YAML files are in %s", self.backup_folder)

    def save_file(self, file_name, file_contents):
        """Save file."""
        self.log.info("Writing file: %s", file_name)
        FileManager.save(file_name, file_contents)


class VersionMigrator:

    """Parent class for a version-specific migrator.

    One instance of this base class exists for each different config file version.
    """

    initialized = False
    deprecations = ''
    renames = ''
    moves = ''
    additions = ''
    migration_logger = None     # type: logging.Logger
    config_version = None       # type: int
    log = logging.getLogger('Migrator')

    def __init__(self, file_name, file_contents):
        """Initialize a specific instance of this file migrator.

        A new instance is created for each file to be migrated.

        Args:
        ----
            file_name: Full path and file name of the file being migrated.
            file_contents: ruamel.load(ed) contents of the file which includes
                the comments.

        Returns a modified file_contents instance with the migrations applied.
        """
        self.log = logging.getLogger(os.path.basename(file_name))
        self.file_name = file_name
        self.base_name = os.path.basename(file_name).lower()
        self.fc = file_contents
        self.current_config_version = 0

        if not self.initialized:
            self._initialize()

        self.log.debug('')
        self.log.debug("------------------ %s ------------------",
                       os.path.basename(file_name))

    @classmethod
    def _initialize(cls):
        # Initializes the class
        cls.migration_logger = logging.getLogger('v%s Migrator' %
                                                 cls.config_version)

        # Deprecations
        cls.deprecations = yaml.load(cls.deprecations)
        try:
            for i, key in enumerate(cls.deprecations):
                cls.deprecations[i] = list(key.split('|'))  # pylint: disable-msg=unsupported-assignment-operation
        except TypeError:
            cls.deprecations = list()

        # Adds
        cls.additions = yaml.load(cls.additions)

        if not cls.additions:
            cls.additions = dict()

        # Renames
        cls.renames = yaml.load(cls.renames)
        try:
            for rename in cls.renames:
                rename['old'] = list(rename['old'].split('|'))
        except TypeError:
            cls.renames = list()

        # Moves
        if cls.moves:
            cls.moves = yaml.load(cls.moves)
            for i, move in enumerate(cls.moves):
                move['old'] = move['old'].split('|')
                move['new'] = move['new'].split('|')
                cls.moves[i] = move     # pylint: disable-msg=unsupported-assignment-operation
        else:
            cls.moves = dict()

        cls.initialized = True

    def migrate(self):
        """Migrate configs and shows."""
        if isinstance(self.fc, CommentedMap):
            if not self._migrate_config_file():
                return False
            self.log.debug("----------------------------------------")
            return self.fc
        if isinstance(self.fc, CommentedSeq):
            if self.is_show_file():
                self._migrate_show_file()
                self.log.debug("----------------------------------------")
                return self.fc

        self.log.debug("Ignoring data file: %s. (Error is ok)", self.file_name)
        return False

    # pylint: disable-msg=no-self-use
    def is_show_file(self):
        """Return true if show file."""
        return False

    def _migrate_show_file(self):
        print(self.file_name, self._get_config_version())

    def _migrate_config_file(self):
        self.log.debug("Analyzing config file: %s", self.file_name)
        if not self._migration_needed():
            return False
        self._update_config_version()
        self.log.debug("Converting keys to lowercase")
        self._do_lowercase()
        self._do_rename()
        self._do_moves()
        self._do_custom()
        self._do_deprecations()
        self._do_adds()
        return True

    def _migration_needed(self):
        self.current_config_version = self._get_config_version()

        if not self.current_config_version:
            self.log.debug("Skipping non-config file: %s", self.file_name)
            return False

        if self.current_config_version == self.config_version:
            if REPROCESS_CURRENT_VERSION:
                self.log.info('Reprocessing file which is already '
                              'config_version=%s', self.config_version)
                return True

            self.log.info('File is already config_version=%s. '
                          'Skipping...', self.config_version)
            return False
        # config_version is less than current:
        if self.config_version - self.current_config_version > 1:
            # use a different loader
            return True
        if self.config_version - self.current_config_version == 1:
            return True

        self.log.warning('MPF version mismatch. File is '
                         'config_version=%s, but this version of'
                         'MPF is for config_version=%s. '
                         'Skipping...',
                         self.current_config_version,
                         self.config_version)
        return False

    def _update_config_version(self):
        # Do a str.replace to preserve any spaces or comments in the header
        self.fc.ca.comment[1][0].value = (
            self.fc.ca.comment[1][0].value.replace(
                'config_version={}'.format(self.current_config_version),
                'config_version={}'.format(self.config_version)))

        self.log.debug('Setting config_version=%s', self.config_version)

    def _get_config_version(self):
        try:
            version_num = int(self.fc.ca.comment[1][0].value.split(
                'config_version=')[1])
            self.log.debug("Current config version is %s", version_num)
            return version_num

        except (TypeError, IndexError):  # No version
            pass

    def _do_deprecations(self):
        # deprecations is a list of lists
        for key in deepcopy(self.deprecations):
            # Everything this does needs the dict, so we pull off the first
            # key so we can get the top level dict
            first_key = key.pop(0)

            if '__list__' in key:
                # If there's a "__list__" string in our list, it means one of
                # the items is a list instead of a dict which we'll have to
                # loop through. In that case, we loop through 20 times, each
                # time checking a different element. That's what the
                # _increment_key_with_list() does.

                key = self._increment_key_with_list(key)

                if first_key not in self.fc:
                    # If the first key isn't even here, we can skip all this
                    continue

                # key found, key is nested
                for dummy_iterator in range(20):

                    if self._remove_key(first_key, key):
                        continue

                    # Not found, but we have a list, so increment & repeat
                    key = self._increment_key_with_list(key)

            else:  # dict only, no list
                self._remove_key(first_key, key)

    def _remove_key(self, first_key, key):
        # actually removes the key from the dict, with nested dicts only
        # (no lists in there)

        if not key:  # single item
            if first_key in self.fc:

                YamlRoundtrip.del_key_with_comments(self.fc, first_key,
                                                    self.log)
                return True

        try:
            if self.fc[first_key].mlget(key, list_ok=True) is not None:
                # mlget just verifies with a nested dict / list
                # index that the key is found.
                final_key = key.pop(-1)
                dic = self.fc[first_key]

                while key:
                    # Loop to get the parent container of the
                    # lowest level key
                    dic = dic[key.pop(0)]

                YamlRoundtrip.del_key_with_comments(dic, final_key, self.log)
                return True
        except (KeyError, IndexError):
            pass

        return False

    @classmethod
    def _increment_key_with_list(cls, key):
        for i, val in enumerate(key):
            if val == '__list__':
                key[i] = 0
            elif isinstance(val, int):
                key[i] += 1

        return key

    def _do_adds(self):
        for section, keys in self.additions.items():
            if section in self.fc:
                for k, v in keys.items():
                    self.log.debug('Adding new key: %s:%s:%s', section, k, v)

                self.fc[section].update(keys)

    def _do_rename(self):
        for rename in self.renames:

            if len(rename['old']) > 1:  # searching for nested key

                found_section = Util.get_from_dict(self.fc, rename['old'][:-1])

                if not found_section:
                    continue

                self.log.debug('Renaming key: %s: -> %s:',
                               ':'.join(rename['old']), rename['new'])
                YamlRoundtrip.rename_key(rename['old'][-1], rename['new'],
                                         found_section)

            else:  # searching for a single key anywhere
                self._recursive_rename(rename['old'][0], rename['new'],
                                       self.fc)

    def _recursive_rename(self, old, new, target):
        if isinstance(target, list):
            for item in target:
                self._recursive_rename(old, new, item)
        elif isinstance(target, dict):

            if old in target:
                YamlRoundtrip.rename_key(old, new, target, self.log)

            for item in target.values():
                self._recursive_rename(old, new, item)

    def _do_moves(self):
        # first move, then rename
        for move in deepcopy(self.moves):
            orig_list = deepcopy(move['old'])
            old_parent = orig_list.pop(0)

            if old_parent not in self.fc:
                continue

            if orig_list:
                old_dict = self.fc[old_parent].mlget(orig_list)
            else:
                old_dict = self.fc[old_parent]

            if not old_dict:
                continue

            new_location = self.fc

            self.log.debug("Moving key: %s -> %s", ':'.join(orig_list),
                           ':'.join(move['new']))

            for key in move['new'][:-1]:
                try:
                    new_location = new_location[key]
                except KeyError:
                    new_location[key] = CommentedMap()
                    new_location = new_location[key]

            new_location[move['new'][-1]] = old_dict
            self._remove_key(old_parent, orig_list)

    def _do_lowercase(self, dic=None):
        # recurcisely converts all keys in dicts and nested dicts
        if not dic:
            dic = self.fc

        key_list = list(dic.keys())

        for key in key_list:
            try:
                YamlRoundtrip.rename_key(key, key.lower(), dic, self.log)
            except AttributeError:
                pass

            try:
                if isinstance(dic[key.lower()], dict):
                    self._do_lowercase(dic[key.lower()])
            except AttributeError:
                if isinstance(dic[key], dict):
                    self._do_lowercase(dic[key])

    def _do_custom(self):
        pass



import json

from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockSocket


class TestFadecandy(MpfTestCase):
    def get_config_file(self):
        return 'fadecandy.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/openpixel/'

    def get_platform(self):
        return 'fadecandy'

    def setUp(self):
        self._messages = []
        super().setUp()
        color_correct = self._messages.pop(0)
        color_correct_binary = b'\x00\xff\x00Z\x00\x01\x00\x01'
        self.assertEqual(color_correct[0:len(color_correct_binary)], color_correct_binary)
        correction = {"linearSlope": 1.0, "linearCutoff": 0.0, "gamma": 2.5, "whitepoint": [1.0, 1.0, 1.0]}
        self.assertEqual(json.loads(color_correct[len(color_correct_binary):].decode()), correction)

        firmware = self._messages.pop(0)
        self.assertEqual(firmware, b'\x00\xff\x00\x05\x00\x01\x00\x02\x00')

        self.assertOpenPixelLedsSent({}, {}, {})

    def tearDown(self):
        self._messages = []
        super().tearDown()
        # test that we blanked all LEDs at shutdown
        self.assertOpenPixelLedsSent({}, {}, {})

    def _mock_loop(self):
        self._mock_socket = MockSocket()
        self.clock.mock_socket("localhost", 7890, self._mock_socket)
        # connect socket to test
        self._mock_socket.send = self._send_mock

    def _build_message(self, channel, leds, max_leds=300):
        out = bytearray()
        out.extend([channel, 0, int(max_leds / 256), max_leds % 256])
        for i in range(0, int(max_leds / 3)):
            if i in leds:
                out.extend([leds[i][0], leds[i][1], leds[i][2]])
            else:
                out.extend([0, 0, 0])

        return bytes(out)

    def _send_mock(self, message):
        self._messages.append(message)
        return len(message)

    def assertOpenPixelLedsSent(self, leds1, leds2, leds3):
        bank1 = self._build_message(0, leds1)
        bank2 = self._build_message(1, leds2)
        bank3 = self._build_message(2, leds3, 9)
        found1 = False
        found2 = False
        found3 = False
        for message in self._messages:
            if message == bank1:
                found1 = True
            elif message == bank2:
                found2 = True
            elif message == bank3:
                found3 = True
            else:
                print("Received:", ":".join("{:02x}".format(c) for c in message))
                print("Expected bank1:", ":".join("{:02x}".format(c) for c in bank1))
                print("Expected bank2:",":".join("{:02x}".format(c) for c in bank2))
                print("Expected bank3:", ":".join("{:02x}".format(c) for c in bank3))
                raise AssertionError("Invalid Message")

        self.assertTrue(found1)
        self.assertTrue(found2)
        self.assertTrue(found3)
        self._messages = []

    def test_led_color(self):
        self.assertEqual("0-63", self.machine.lights["test_led_serial"].hw_drivers["red"][0].number)
        self.assertEqual("0-64", self.machine.lights["test_led_serial"].hw_drivers["green"][0].number)
        self.assertEqual("0-65", self.machine.lights["test_led_serial"].hw_drivers["blue"][0].number)
        self.assertEqual("0-66", self.machine.lights["test_led_serial2"].hw_drivers["red"][0].number)
        self.assertEqual("0-67", self.machine.lights["test_led_serial2"].hw_drivers["green"][0].number)
        self.assertEqual("0-68", self.machine.lights["test_led_serial2"].hw_drivers["blue"][0].number)
        self.assertEqual("0-69", self.machine.lights["test_led_serial2"].hw_drivers["white"][0].number)

        # test led on channel 0. position 99
        self.machine.lights["test_led"].on()
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, {}, {})

        # test led 20 ond channel 0
        self.machine.lights["test_led2"].color(RGBColor((255, 0, 0)))
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (255, 255, 255)}, {}, {})

        self.machine.lights["test_led"].off()
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (0, 0, 0)}, {}, {})
        self._messages = []

        # test led color
        self.machine.lights["test_led"].color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (2, 23, 42)}, {}, {})

        # test led on channel 1
        self.machine.lights["test_led3"].on()
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (2, 23, 42)}, {99: (255, 255, 255)}, {})

        self.machine.lights["test_led"].off()
        self.machine.lights["test_led2"].off()
        self.machine.lights["test_led3"].off()
        self.advance_time_and_run(.02)
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({}, {}, {})
        self.machine.lights["test_led"].color("blue", priority=5, fade_ms=20, key="test2")
        self.advance_time_and_run()

        self.machine.lights["test_led"].on(priority=10, fade_ms=1000, key="test")
        self.advance_time_and_run(2)
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, {}, {})
        # test crossfade on same key
        self._messages = []
        self.machine.lights["test_led"].on(priority=10, fade_ms=1000, key="test")
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, {}, {})

        # test remove and readd
        self._messages = []
        self.machine.lights["test_led"].remove_from_stack_by_key(fade_ms=1000, key="test")
        self.machine.lights["test_led"].on(priority=10, fade_ms=1000, key="test")
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, {}, {})

        # test fade out
        self.machine.lights["test_led"].remove_from_stack_by_key(fade_ms=1000, key="test")
        self.advance_time_and_run(.5)
        self._messages = []
        self.advance_time_and_run(.1)
        found = False
        for message in self._messages:
            if message[0] == 0 and len(message) >= 304:
                if 90 < message[301] < 100 and 90 < message[302] < 100 and message[303] == 255:
                    found = True
        self.assertTrue(found)
        self.advance_time_and_run(.5)
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({99: (0, 0, 255)}, {}, {})

        # fade over fade
        self.machine.lights["test_led"].on(priority=10, key="test")
        self.machine.lights["test_led"].remove_from_stack_by_key(fade_ms=10000, key="test2")
        self.advance_time_and_run(.1)
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, {}, {})
        self.machine.lights["test_led"].remove_from_stack_by_key(fade_ms=1000, key="test")
        self.advance_time_and_run(.5)
        self.advance_time_and_run(20)
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({}, {}, {})

        # set RGBW
        self._messages = []
        self.machine.lights["test_rgbw"].on(priority=10, key="test")
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({}, {}, {0: (255, 255, 255), 1: (0, 255, 0)})

        self._messages = []
        self.machine.lights["test_rgbw"].color("red", priority=20, key="test")
        self._messages = []
        self.advance_time_and_run(.1)
        self.assertOpenPixelLedsSent({}, {}, {0: (0, 255, 0), 1: (0, 0, 0)})



"""Interface for monochrome and rgb platform devices."""
from typing import List

import abc


class DmdPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for monochrome DMDs in hardware platforms."""

    __slots__ = []  # type: List[str]

    @abc.abstractmethod
    def update(self, data: bytes):
        """Update data on the DMD.

        Args:
        ----
            data: bytes to send to DMD
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_brightness(self, brightness: float):
        """Set brightness of DMD."""
        raise NotImplementedError



"""Contains code for a virtual hardware platform."""
from typing import Dict, Tuple, Optional, Union

import asyncio
import logging

from mpf.platforms.interfaces.hardware_sound_platform_interface import HardwareSoundPlatformInterface
from mpf.platforms.interfaces.i2c_platform_interface import I2cPlatformInterface
from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplayPlatformInterface, FlashingType

from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface

from mpf.core.platform import ServoPlatform, SwitchPlatform, DriverPlatform, AccelerometerPlatform, I2cPlatform, \
    DmdPlatform, RgbDmdPlatform, LightsPlatform, DriverConfig, SwitchConfig, SegmentDisplayPlatform, StepperPlatform, \
    HardwareSoundPlatform, SwitchSettings, DriverSettings, RepulseSettings
from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings


class VirtualHardwarePlatform(AccelerometerPlatform, I2cPlatform, ServoPlatform, LightsPlatform, SwitchPlatform,
                              DriverPlatform, DmdPlatform, RgbDmdPlatform, SegmentDisplayPlatform, StepperPlatform,
                              HardwareSoundPlatform):

    """Base class for the virtual hardware platform."""

    __slots__ = ["hw_switches", "initial_states_sent", "_next_driver", "_next_switch", "_next_light", "__dict__",
                 "rules"]

    def __init__(self, machine) -> None:
        """Initialise virtual platform."""
        super().__init__(machine)
        self._setup_log()

        # Since the virtual platform doesn't have real hardware, we need to
        # maintain an internal list of switches that were confirmed so we have
        # something to send when MPF needs to know the hardware states of
        # switches
        self.hw_switches = dict()   # type: Dict[str, bool]
        self.initial_states_sent = False
        self.features['tickless'] = True
        self.features['allow_empty_numbers'] = True
        self._next_driver = 1000
        self._next_switch = 1000
        self._next_light = 1000
        self.rules = {}     # type: Dict[Tuple[SwitchPlatformInterface, DriverPlatformInterface], str]

    def __repr__(self):
        """Return string representation."""
        return '<Platform.Virtual>'

    def _setup_log(self):
        self.log = logging.getLogger("Virtual Platform")
        self.log.debug("Configuring virtual hardware interface.")

    async def initialize(self) -> None:
        """Initialise platform."""

    def stop(self):
        """Stop platform."""

    async def configure_servo(self, number: str):
        """Configure a servo device in platform."""
        return VirtualServo(number)

    async def configure_stepper(self, number: str, config: dict):
        """Configure a smart stepper / axis device in platform."""
        del config
        return VirtualStepper(number, self.machine)

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure driver."""
        del platform_settings
        # generate number if None
        if number is None:
            number = str(self._next_driver)
            self._next_driver += 1

        driver = VirtualDriver(config, number)

        return driver

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure switch."""
        # switch needs a number to be distingishable from other switches
        if number is None:
            number = self._next_switch
            self._next_switch += 1

        # We want to have the virtual platform set all the initial switch states
        # to inactive, so we have to check the config.
        self.hw_switches[number] = config.invert

        return VirtualSwitch(config, number)

    async def get_hw_switch_states(self):
        """Return hw switch states."""
        if not self.initial_states_sent:

            if 'virtual_platform_start_active_switches' in self.machine.config:
                initial_active_switches = []
                for switch in Util.string_to_list(self.machine.config['virtual_platform_start_active_switches']):
                    if switch not in self.machine.switches:
                        if " " in switch:
                            self.raise_config_error("MPF no longer supports lists separated by space in "
                                                    "virtual_platform_start_active_switches. Please separate "
                                                    "switches by comma: {}.".format(switch), 1)
                        else:
                            self.raise_config_error("Switch {} used in virtual_platform_start_active_switches was not "
                                                    "found in switches section.".format(switch), 1)
                    initial_active_switches.append(self.machine.switches[switch].hw_switch.number)

                for k in self.hw_switches:
                    if k in initial_active_switches:
                        self.hw_switches[k] ^= 1

            self.initial_states_sent = True

        else:
            switches = [x for x in self.machine.switches.values() if x.platform == self]

            for switch in switches:
                self.hw_switches[switch.hw_switch.number] = switch.state ^ switch.invert

        return self.hw_switches

    def _get_platforms(self):
        platforms = []
        for name, platform in self.machine.config['mpf']['platforms'].items():
            if name in ("virtual", "smart_virtual"):
                continue
            platforms.append(Util.string_to_class(platform))
        return platforms

    def validate_stepper_section(self, stepper, config):
        """Validate stepper sections."""
        return config

    def validate_switch_section(self, switch, config):
        """Validate switch sections."""
        return config

    def validate_coil_section(self, driver, config):
        """Validate coil sections."""
        return config

    def configure_accelerometer(self, number, config, callback):
        """Configure accelerometer."""

    def configure_light(self, number, subtype, config, platform_settings):
        """Configure light channel."""
        del config
        if not subtype:
            subtype = "led"
        return VirtualLight("{}-{}".format(subtype, number), platform_settings, self.machine)

    # pylint: disable-msg=no-self-use
    def configure_hardware_sound_system(self) -> "HardwareSoundPlatformInterface":
        """Configure virtual hardware sound system."""
        return VirtualSound()

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse channel str to a list of channels."""
        if number is None:
            number = self._next_light
            self._next_light += 1
        if subtype in ("gi", "matrix"):
            return [
                {
                    "number": str(number)
                }
            ]
        if subtype == "led" or not subtype:
            return [
                {
                    "number": str(number) + "-r",
                },
                {
                    "number": str(number) + "-g",
                },
                {
                    "number": str(number) + "-b",
                }
            ]

        raise AssertionError("Unknown subtype {}".format(subtype))

    def clear_hw_rule(self, switch, coil):
        """Clear hw rule."""
        if (switch.hw_switch, coil.hw_driver) in self.rules:
            del self.rules[(switch.hw_switch, coil.hw_driver)]
        else:
            self.log.debug("Tried to clear a non-existing rules %s <-> %s", switch, coil)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch, coil):
        """Set rule."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_enable_and_release"

    def set_pulse_on_hit_and_release_rule(self, enable_switch, coil):
        """Set rule."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_release"

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set rule."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_release_and_disable"

        if (eos_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                eos_switch.hw_switch, coil.hw_driver))

        self.rules[(eos_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_release_and_disable"

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings,
                                                                 coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set rule."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_enable_and_release_and_disable"

        if (eos_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                eos_switch.hw_switch, coil.hw_driver))

        self.rules[(eos_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit_and_enable_and_release_and_disable"

    def set_pulse_on_hit_rule(self, enable_switch, coil):
        """Set rule."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = "pulse_on_hit"

    def configure_dmd(self):
        """Configure DMD."""
        return VirtualDmd()

    def configure_rgb_dmd(self, name: str):
        """Configure DMD."""
        del name
        return VirtualDmd()

    async def configure_segment_display(self, number: str, platform_settings) -> SegmentDisplayPlatformInterface:
        """Configure segment display."""
        del platform_settings
        return VirtualSegmentDisplay(number)

    async def configure_i2c(self, number: str) -> "I2cPlatformInterface":
        """Configure virtual i2c device."""
        return VirtualI2cDevice(number, self._get_initial_i2c(number))

    @staticmethod
    def _get_initial_i2c(number):
        """Get virtual i2c layout.

        Mock this in your test.
        """
        del number
        return {}


class VirtualI2cDevice(I2cPlatformInterface):

    """Virtual i2c device."""

    __slots__ = ["data"]

    def __init__(self, number, initial_layout) -> None:
        """Initialise virtual i2c device."""
        super().__init__(number)
        self.data = initial_layout

    def i2c_write8(self, register, value):
        """Write data."""
        self.data[int(register)] = value

    async def i2c_read_block(self, register, count):
        """Read data block."""
        result = []
        for i in range(int(register), int(register) + count):
            result.append(self.data[i])
        return result

    async def i2c_read8(self, register):
        """Read data."""
        return self.data[int(register)]


class VirtualSegmentDisplay(SegmentDisplayPlatformInterface):

    """Virtual segment display."""

    __slots__ = ["text", "flashing"]

    def __init__(self, number) -> None:
        """Initialise virtual segment display."""
        super().__init__(number)
        self.text = ''
        self.flashing = FlashingType.NO_FLASH

    def set_text(self, text: str, flashing: FlashingType):
        """Set text."""
        self.text = text
        self.flashing = flashing


class VirtualSound(HardwareSoundPlatformInterface):

    """Virtual hardware sound interface."""

    __slots__ = ["playing", "volume"]

    def __init__(self) -> None:
        """Initialise virtual hardware sound."""
        self.playing = None     # type: Optional[Union[int, str]]
        self.volume = None      # type: Optional[float]

    def play_sound(self, number: int, track: int = 1):
        """Play virtual sound."""
        self.playing = number

    def play_sound_file(self, file: str, platform_options: dict, track: int = 1):
        """Play a sound file."""
        self.playing = file

    def text_to_speech(self, text: str, platform_options: dict, track: int = 1):
        """Text to speech output."""
        self.playing = text

    def set_volume(self, volume: float, track: int = 1):
        """Set volume."""
        self.volume = volume

    def stop_all_sounds(self, track: int = 1):
        """Stop sound."""
        self.playing = None


class VirtualDmd(DmdPlatformInterface):

    """Virtual DMD."""

    __slots__ = ["data", "brightness"]

    def __init__(self) -> None:
        """Initialise virtual DMD."""
        self.data = None        # type: Optional[bytes]
        self.brightness = None  # type: Optional[float]

    def update(self, data: bytes):
        """Update data on the DMD.

        Args:
        ----
            data: bytes to send to DMD
        """
        self.data = data

    def set_brightness(self, brightness: float):
        """Set brightness."""
        self.brightness = brightness


class VirtualSwitch(SwitchPlatformInterface):

    """Represents a switch in a pinball machine used with virtual hardware."""

    __slots__ = ["log"]

    def __init__(self, config, number) -> None:
        """Initialise switch."""
        super().__init__(config, number)
        self.log = logging.getLogger('VirtualSwitch')

    def get_board_name(self):
        """Return the name of the board of this switch."""
        return "Virtual"

    def __repr__(self):
        """Str representation."""
        return "VirtualSwitch.{}".format(self.number)


class VirtualLight(LightPlatformInterface):

    """Virtual Light."""

    __slots__ = ["settings", "_current_fade", "machine"]

    def __init__(self, number, settings, machine) -> None:
        """Initialise LED."""
        super().__init__(number)
        self.settings = settings
        self.machine = machine
        self._current_fade = (0, -1, 0, -1)

    @property
    def current_brightness(self) -> float:
        """Return current brightness."""
        current_time = self.machine.clock.get_time()
        start_brightness, start_time, target_brightness, target_time = self._current_fade
        if target_time > current_time:
            ratio = ((current_time - start_time) /
                     (target_time - start_time))
            return start_brightness + (target_brightness - start_brightness) * ratio

        return target_brightness

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Store CB function."""
        self._current_fade = (start_brightness, start_time, target_brightness, target_time)

    def get_board_name(self):
        """Return the name of the board of this light."""
        return "Virtual"

    def is_successor_of(self, other):
        """Return true if the other light has the same number string plus the suffix '+1'."""
        return self.number == other.number + "+1"

    def get_successor_number(self):
        """Return the number with the suffix '+1'.

        As there is not real number format for virtual is this is all we can do here.
        """
        return self.number + "+1"

    def __lt__(self, other):
        """Order lights by string."""
        return self.number < other.number


class VirtualServo(ServoPlatformInterface):

    """Virtual servo."""

    __slots__ = ["log", "number", "current_position", "speed_limit", "acceleration_limit"]

    def __init__(self, number) -> None:
        """Initialise servo."""
        self.log = logging.getLogger('VirtualServo')
        self.number = number
        self.current_position = None
        self.speed_limit = None
        self.acceleration_limit = None

    def go_to_position(self, position):
        """Go to position."""
        self.current_position = position

    def set_speed_limit(self, speed_limit):
        """Set speed parameter."""
        self.speed_limit = speed_limit

    def set_acceleration_limit(self, acceleration_limit):
        """Set acceleration parameter."""
        self.acceleration_limit = acceleration_limit


class VirtualStepper(StepperPlatformInterface):

    """Virtual Stepper."""

    __slots__ = ["log", "number", "_current_position", "velocity", "direction", "machine"]

    def __init__(self, number, machine) -> None:
        """Initialise servo."""
        self.log = logging.getLogger('VirtualStepper')
        self.number = number
        self._current_position = 0
        self.velocity = 0
        self.direction = 0  # clockwise
        self.machine = machine

    def home(self, direction):
        """Home an axis, resetting 0 position."""
        self._current_position = 0

    async def wait_for_move_completed(self):
        """Wait until move completed."""
        await asyncio.sleep(0.1)

    def move_rel_pos(self, position):
        """Move axis to a relative position."""
        self._current_position += position

    def move_vel_mode(self, velocity):
        """Move at a specific velocity indefinitely."""
        self.velocity = velocity

    def current_position(self):
        """Return current position of stepper."""
        return self._current_position

    def stop(self):
        """Stop motor."""
        self.velocity = 0


class VirtualDriver(DriverPlatformInterface):

    """A virtual driver object."""

    __slots__ = ["state", "log", "__dict__"]

    def __init__(self, config, number) -> None:
        """Initialise virtual driver to disabled."""
        self.log = logging.getLogger("VirtualDriver.{}".format(number))
        super().__init__(config, number)
        self.state = "disabled"

    def get_board_name(self):
        """Return the name of the board of this driver."""
        return "Virtual"

    def __repr__(self):
        """Str representation."""
        return "VirtualDriver.{}".format(self.number)

    def disable(self):
        """Disable virtual coil."""
        self.log.debug("Disabling driver")
        self.state = "disabled"

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable virtual coil."""
        del pulse_settings, hold_settings
        self.log.debug("Enabling driver")
        self.state = "enabled"

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse virtual coil."""
        self.log.debug("Pulsing driver for %sms", pulse_settings.duration)
        self.state = "pulsed_" + str(pulse_settings.duration)



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestMagnet(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/magnet/'

    def test_enable_on_game_start(self):
        self.assertFalse(self.machine.magnets["magnet_auto_enable"].enabled)
        self.mock_event("magnet_magnet_auto_enable_grabbing_ball")
        self.mock_event("magnet_magnet_auto_enable_grabbed_ball")
        self.assertEqual("disabled", self.machine.coils["magnet_coil3"].hw_driver.state)
        # hit switch
        self.hit_and_release_switch("grab_switch3")
        self.advance_time_and_run(.1)
        # magnet not enabled (game not started yet) -> nothing happens
        self.assertEqual("disabled", self.machine.coils["magnet_coil3"].hw_driver.state)
        # start game
        self.start_game()
        self.assertTrue(self.machine.magnets["magnet_auto_enable"].enabled)

        # try again
        self.hit_and_release_switch("grab_switch3")
        self.advance_time_and_run(.1)

        # magnet enables
        self.assertEventCalled("magnet_magnet_auto_enable_grabbing_ball")
        self.assertEventNotCalled("magnet_magnet_auto_enable_grabbed_ball")
        self.assertEqual("enabled", self.machine.coils["magnet_coil3"].hw_driver.state)

    def test_grab_and_release_and_fling(self):
        self.mock_event("magnet_magnet1_grabbing_ball")
        self.mock_event("magnet_magnet1_grabbed_ball")
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        # hit switch
        self.hit_and_release_switch("grab_switch1")
        self.advance_time_and_run(.1)
        # magnet not enabled -> nothing happens
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        # enable magnet and try again
        self.post_event("magnet1_enable")
        self.hit_and_release_switch("grab_switch1")
        self.advance_time_and_run(.1)

        # magnet enables
        self.assertEventCalled("magnet_magnet1_grabbing_ball")
        self.assertEventNotCalled("magnet_magnet1_grabbed_ball")
        self.assertEqual("enabled", self.machine.coils["magnet_coil1"].hw_driver.state)

        # success event after 1.5s
        self.advance_time_and_run(1.5)
        self.assertEventCalled("magnet_magnet1_grabbing_ball")
        self.assertEventCalled("magnet_magnet1_grabbed_ball")

        self.mock_event("magnet_magnet1_releasing_ball")
        self.mock_event("magnet_magnet1_released_ball")
        # release ball
        self.post_event("magnet1_release")
        self.advance_time_and_run(.1)
        # coil disables
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        self.assertEventCalled("magnet_magnet1_releasing_ball")
        self.assertEventNotCalled("magnet_magnet1_released_ball")

        # ball passes switch
        self.hit_and_release_switch("grab_switch1")

        # release done after .5s
        self.advance_time_and_run(.5)
        self.assertEventCalled("magnet_magnet1_releasing_ball")
        self.assertEventCalled("magnet_magnet1_released_ball")
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)

        # grab again
        self.mock_event("magnet_magnet1_grabbing_ball")
        self.mock_event("magnet_magnet1_grabbed_ball")
        self.hit_and_release_switch("grab_switch1")
        self.advance_time_and_run(.1)

        # magnet enables
        self.assertEventCalled("magnet_magnet1_grabbing_ball")
        self.assertEventNotCalled("magnet_magnet1_grabbed_ball")
        self.assertEqual("enabled", self.machine.coils["magnet_coil1"].hw_driver.state)

        # success event after 1.5s
        self.advance_time_and_run(1.5)
        self.assertEventCalled("magnet_magnet1_grabbing_ball")
        self.assertEventCalled("magnet_magnet1_grabbed_ball")

        # fling ball
        self.mock_event("magnet_magnet1_flinging_ball")
        self.mock_event("magnet_magnet1_flinged_ball")
        self.post_event("magnet1_fling")
        self.advance_time_and_run(.01)
        # coil disables
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        self.assertEventCalled("magnet_magnet1_flinging_ball")
        self.assertEventNotCalled("magnet_magnet1_flinged_ball")

        # coil reenables
        self.advance_time_and_run(.25)
        self.assertEqual("enabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        self.assertEventCalled("magnet_magnet1_flinging_ball")
        self.assertEventNotCalled("magnet_magnet1_flinged_ball")

        # and disables again
        self.advance_time_and_run(.1)
        self.assertEqual("disabled", self.machine.coils["magnet_coil1"].hw_driver.state)
        self.assertEventCalled("magnet_magnet1_flinging_ball")
        self.assertEventCalled("magnet_magnet1_flinged_ball")

    def test_magnet_ball_save(self):
        # enable magnet ball save
        self.post_event("magnet_ball_save_enable")
        self.assertFalse(self.machine.ball_saves["magnet_save"].enabled)
        # ball passes flipper fingers and wants to drain
        self.hit_and_release_switch("grab_switch2")
        self.advance_time_and_run(.01)
        # but the magnet enables and saves it
        self.assertEqual("enabled", self.machine.coils["magnet_coil2"].hw_driver.state)
        self.assertTrue(self.machine.ball_saves["magnet_save"].enabled)
        # after 1.5s the magnet flings the ball up
        self.advance_time_and_run(1.5)
        self.assertEqual("disabled", self.machine.coils["magnet_coil2"].hw_driver.state)

        # coil reenables
        self.advance_time_and_run(.25)
        self.assertEqual("enabled", self.machine.coils["magnet_coil2"].hw_driver.state)

        # and disables again
        self.advance_time_and_run(.1)
        self.assertEqual("disabled", self.machine.coils["magnet_coil2"].hw_driver.state)

        # ball save stays enabled for a short moment in case the magnet missed it
        self.advance_time_and_run(3)
        self.assertTrue(self.machine.ball_saves["magnet_save"].enabled)

        # game continues and ball save times out
        self.advance_time_and_run(5)
        self.assertFalse(self.machine.ball_saves["magnet_save"].enabled)



"""Base class for all bcp clients."""
import abc

from mpf.core.mpf_controller import MpfController


class BaseBcpClient(MpfController, metaclass=abc.ABCMeta):

    """Base class for bcp clients."""

    __slots__ = ["name", "bcp", "exit_on_close"]

    def __init__(self, machine, name, bcp):
        """Initialise client."""
        super().__init__(machine)
        self.name = name
        self.bcp = bcp
        self.exit_on_close = False

    async def connect(self, config):
        """Actively connect client."""
        raise NotImplementedError("implement")

    async def read_message(self):
        """Read one message from client."""
        raise NotImplementedError("implement")

    def accept_connection(self, receiver, sender):
        """Handle incoming connection from remote client."""
        raise NotImplementedError("implement")

    def send(self, bcp_command, kwargs):
        """Send data to client."""
        raise NotImplementedError("implement")

    def stop(self):
        """Stop client connection."""
        raise NotImplementedError("implement")



from mpf.tests.MpfTestCase import MpfTestCase


class TestPlayfield(MpfTestCase):

    def get_config_file(self):
        return 'test_playfield.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/playfield/'

    # nothing to test currently



"""A system wide device which can be defined in the main config."""
from typing import List

import abc

from mpf.core.device import Device


class SystemWideDevice(Device, metaclass=abc.ABCMeta):

    """A system wide device which can be defined in the main config."""

    __slots__ = []  # type: List[str]

    async def device_added_system_wide(self):
        """Add the device system wide."""
        await self._initialize()



"""Flasher config player."""
from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.core.delays import DelayManager
from mpf.core.utility_functions import Util


class FlasherPlayer(DeviceConfigPlayer):

    """Triggers flashers based on config."""

    config_file_section = 'flasher_player'
    show_section = 'flashers'

    __slots__ = ["delay"]

    def __init__(self, machine):
        """Initialise flasher_player."""
        super().__init__(machine)
        self.delay = DelayManager(self.machine)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Flash flashers."""
        del kwargs

        for flasher, s in settings.items():
            if isinstance(flasher, str):
                flasher_names = Util.string_to_event_list(flasher)
                for flasher_name in flasher_names:
                    self._flash(self.machine.lights[flasher_name],
                                duration_ms=s['ms'],
                                color=s['color'],
                                key=context)
            else:
                self._flash(flasher, duration_ms=s['ms'], key=context, color=s['color'])

    def _flash(self, light, duration_ms, key, color):
        light.color(color, fade_ms=0, key=key)
        self.delay.add(duration_ms, self._remove_flash, light=light, key=key)

    @staticmethod
    def _remove_flash(light, key):
        light.remove_from_stack_by_key(key=key, fade_ms=0)

    def get_express_config(self, value):
        """Parse express config."""
        return dict(ms=value)
